"""Parameters.
==============

This module defines a utility class to store and manage neural net parameters and metadata."""

import json
import os
from dataclasses import dataclass
from pathlib import Path
from typing import Iterable, List, Union

import jsonpointer
import jsonschema
import numpy as np
import orjson

from .activation import ACTIVATIONS

_here = Path(os.path.dirname(os.path.abspath(__file__)))
SCHEMA = json.loads((_here / "schema.json").read_text())


@dataclass
class Parameters:
    r"""Neural network parameters.

    .. warning::
        The attributes of this class are not protected. It's possible
        to overwrite them instead of updating them in place. To ensure
        that an array is updated in place, use the numpy `[:]` syntax:

        .. code-block:: python

            parameters = Parameters(**kwargs)
            layer_1_weights = parameters.W[1]
            layer_1_weights[:] = new_array_values  # note [:]

    .. note::

        The variables and their symbols refer to the theory in the companion
        `paper`_ for this library.

    :param layer_sizes: number of nodes in each layer (including
        input/output layers)
    :param hidden_activation: activation function used in hidden layers
    :param output_activation: activation function used in output layer

    :ivar W: weights :math:`\boldsymbol{W} \in \mathbb{R}^{n^{[l]} \times n^{[l-1]}}` for each layer
    :vartype W: List[np.ndarray]

    :ivar b: biases :math:`\boldsymbol{b} \in \mathbb{R}^{n^{[l]} \times 1}` for each layer
    :vartype b: List[np.ndarray]

    :ivar a: activation names for each layer
    :vartype a: List[str]

    :ivar dW: partials w.r.t. weight :math:`dL/dW^{[l]} \in \mathbb{R}^{n^{[l]} \times n^{[l-1]}}`
    :vartype dW: List[np.ndarray]

    :ivar db: partials w.r.t. bias :math:`dL/db^{[l]} \in \mathbb{R}^{n^{[l]} \times 1}`
    :vartype db: List[np.ndarray]

    :ivar mu_x: mean of training data inputs used for normalization :math:`\mu_x \in \mathbb{R}^{n_x \times 1}`
    :vartype mu_x: List[np.ndarray]

    :ivar mu_y: mean of training data outputs used for normalization :math:`\mu_y \in \mathbb{R}^{n_y \times 1}`
    :vartype mu_x: List[np.ndarray]

    :ivar sigma_x: standard deviation of training data inputs used for normalization :math:`\sigma_x \in \mathbb{R}^{n_x \times 1}`
    :vartype sigma_x: List[np.ndarray]

    :ivar sigma_y: standard deviation of training data outputs used for normalization :math:`\sigma_y \in \mathbb{R}^{n_y \times 1}`
    :vartype sigma_y: List[np.ndarray]
    """

    layer_sizes: List[int]
    hidden_activation: str = "tanh"
    output_activation: str = "linear"

    @property
    def layers(self) -> Iterable[int]:
        """Return iterator of index for each layer."""
        return range(self.L)

    @property
    def partials(self) -> Iterable[int]:
        """Return iterator of index for each partial."""
        return range(self.n_x)

    @property
    def n_x(self) -> int:
        """Return number of inputs."""
        return self.layer_sizes[0]

    @property
    def n_y(self) -> int:
        """Return number of outputs."""
        return self.layer_sizes[-1]

    @property
    def L(self) -> int:
        """Return number of layers."""
        return len(self.layer_sizes)

    def initialize(self, random_state: Union[int, None] = None) -> None:
        """Use `He initialization <https://arxiv.org/pdf/1502.01852.pdf>`_ to
        initialize parameters.

        :param random_state: optional random seed (for repeatability)
        """
        rng = np.random.default_rng(random_state)
        self.W = []
        self.b = []
        self.a = []
        self.dW = []
        self.db = []
        self.mu_x = np.zeros((self.n_x, 1))
        self.mu_y = np.zeros((self.n_y, 1))
        self.sigma_x = np.ones((self.n_x, 1))
        self.sigma_y = np.ones((self.n_y, 1))
        previous_layer_size = -1  # Not used on first loop.
        for i, layer_size in enumerate(self.layer_sizes):
            if i == 0:  # input layer
                W = np.eye(layer_size)
                b = np.zeros((layer_size, 1))
                a = "linear"
            elif i == self.L - 1:  # output layer
                W = rng.normal(size=(layer_size, previous_layer_size)) * np.sqrt(
                    1.0 / previous_layer_size
                )
                b = np.zeros((layer_size, 1))
                a = self.output_activation
            else:  # hidden layer
                W = rng.normal(size=(layer_size, previous_layer_size)) * np.sqrt(
                    1.0 / previous_layer_size
                )
                b = np.zeros((layer_size, 1))
                a = self.hidden_activation
            dW = np.zeros(W.shape)
            db = np.zeros(b.shape)
            self.dW.append(dW)
            self.db.append(db)
            self.W.append(W)
            self.b.append(b)
            self.a.append(a)
            previous_layer_size = layer_size

    def stack(self) -> np.ndarray:
        """Stack W, b into a single array.

        .. code-block::

            parameters.stack()
            >> np.array([[W1], [b1], [W2], [b2], [W3], [b3]])

        .. note::
            This method is used to convert the list format
            used by the neural net into a single array of stacked parameters
            for optimization.
        """
        stacks = self.stack_per_layer()
        return np.concatenate(stacks).reshape((-1, 1))

    def stack_per_layer(self) -> List[np.ndarray]:
        """Stack W, b into a single array for each layer.

        .. code-block::

            parameters.stack_per_layer()
            >> [np.array([[W1], [b1]]), [W2], [b2]]), np.array([[W3], [b3]])]
        """
        stacks = []
        for i in range(self.L):
            stack = np.concatenate([self.W[i].ravel(), self.b[i].ravel()]).reshape(
                (-1, 1)
            )
            stacks.append(stack)
        return stacks

    def stack_partials(self) -> np.ndarray:
        """Stack backprop partials dW, db.

        .. code-block::

            parameters.stack_partials()
            >> np.array([[dW1], [db1], [dW2], [db2], [dW3], [db3]])

        .. note::
            This method is used to convert the list format used by the neural
            net into a single array of stacked parameters for optimization.
        """
        stacks = self.stack_partials_per_layer()
        return np.concatenate(stacks).reshape((-1, 1))

    def stack_partials_per_layer(self) -> List[np.ndarray]:
        """Stack backprop partials dW, db per layer.

        .. code-block::

            parameters.stack_partials_per_layer()
            >> [np.array([[dW1], [db1]]), np.array([[dW2], [db2]]), np.array([[dW3], [db3]]),]
        """
        stacks = []
        for i in range(self.L):
            stack = np.concatenate(
                [
                    self.dW[i].ravel(),
                    self.db[i].ravel(),
                ]
            ).reshape((-1, 1))
            stacks.append(stack)
        return stacks

    def _column_to_stacks(self, params: np.ndarray) -> List[np.ndarray]:
        """Convert parameters from single stack to list of stacks.

        Neural net parameters are converted from single stack
        representation (for all layers) to a list of stacks (per layer).

        Parameters
        ----------
        params: np.ndarray
            Neural network parameters as single array where all layers
            are stacked on top of each other.
            e.g. np.array([[W1], [b1], [W2], [b2], [W3], [b3]])

        Returns
        -------
        params: List[np.ndarray]
            List of stacks (one per layer)
            e.g. [np.array([[W1], [b1]]), [W2], [b2]]), np.array([[W3], [b3]])]
        """
        stacks = []
        k = 0
        for i in range(self.L):  # single stack to many stacks (for each layer)
            n_w, p = self.W[i].shape
            n_b, _ = self.b[i].shape
            n = n_w * p + n_b
            stack = params[k : k + n]
            stacks.append(stack)
            k += n
        return stacks

    def unstack(self, parameters: Union[np.ndarray, List[np.ndarray]]) -> None:
        """Unstack parameters W, b back into list of arrays.

        :param parameters: neural network parameters as either a single
            array where all layers are stacked on top of each other or a list of
            stacked parameters for each layer.

        .. code-block::

            # Unstack from single stack
            parameters.unstack(np.array([[W1], [b1], [W2], [b2], [W3], [b3]]))
            parameters.W, parameters.b
            >> [W1, W2, W3], [b1, b2, b3]

            # Unstack from list of stacks
            parameters.unstack([np.array([[W1], [b1]]), [W2], [b2]]), np.array([[W3], [b3]])])
            parameters.W, parameters.b
            >> [W1, W2, W3], [b1, b2, b3]

        .. note::
            This method is used to convert optimization results expressed
            as a single array of stacked parameters, back into the list format
            used by the neural net.
        """
        if isinstance(parameters, np.ndarray):  # single column
            parameters = self._column_to_stacks(parameters)
        for i, array in enumerate(parameters):  # stacks to params for each layer
            n, p = self.W[i].shape
            self.W[i][:] = array[: n * p].reshape(n, p)
            self.b[i][:] = array[n * p :].reshape(n, 1)

    def unstack_partials(self, partials: Union[np.ndarray, List[np.ndarray]]) -> None:
        """Unstack backprop partials dW, db back into list of arrays.

        :param partials: neural network partials as either a single
            array where all layers are stacked on top of each other or a list of
            stacked parameters for each layer.

        .. code-block::

            # Unstack from single stack
            parameters.unstack(np.array([[dW1], [db1], [dW2], [db2], [dW3], [db3]]))
            parameters.dW, parameters.db
            >> [dW1, dW2, dW3], [db1, db2, db3]

            # Unstack from list of stacks
            parameters.unstack([np.array([[dW1], [db1]]), [dW2], [db2]]), np.array([[dW3], [db3]])])
            parameters.dW, parameters.db
            >> [dW1, dW2, dW3], [db1, db2, db3]

        .. note::
            This method is used to convert optimization results expressed
            as a single array of stacked parameters, back into the list format
            used by the neural net.
        """
        if isinstance(partials, np.ndarray):  # single column
            partials = self._column_to_stacks(partials)
        for i, array in enumerate(partials):
            n, p = self.dW[i].shape
            self.dW[i][:] = array[: n * p].reshape(n, p)
            self.db[i][:] = array[n * p :].reshape(n, 1)

    def _serialize(self) -> bytes:
        """Serialize parameters into byte stream for json."""
        keys = jsonpointer.JsonPointer("/properties").get(SCHEMA)
        data = {key: getattr(self, key) for key in keys}
        return orjson.dumps(data, option=orjson.OPT_SERIALIZE_NUMPY)

    def _deserialize(self, saved_parameters: bytes) -> None:
        """Deserialize and apply saved parameters."""
        params = orjson.loads(saved_parameters)
        jsonschema.validate(params, SCHEMA)
        self.W = [np.array(value) for value in params["W"]]
        self.b = [np.array(value) for value in params["b"]]
        self.a = params["a"]
        self.mu_x = np.array(params["mu_x"])
        self.mu_y = np.array(params["mu_y"])
        self.sigma_x = np.array(params["sigma_x"])
        self.sigma_y = np.array(params["sigma_y"])
        self.layer_sizes = [W.shape[0] for W in self.W]
        self.output_activation = self.a[-1]
        self.hidden_activation = self.a[-2]
        self.dW = [np.zeros(array.shape) for array in self.W]
        self.db = [np.zeros(array.shape) for array in self.b]
        assert (
            self.mu_x.size == self.layer_sizes[0]
        ), "mu_x size is different input layer size"
        assert (
            self.mu_y.size == self.layer_sizes[-1]
        ), "mu_y size is different output layer size"
        assert (
            self.sigma_x.size == self.layer_sizes[0]
        ), "sigma_x size is different input layer size"
        assert (
            self.sigma_y.size == self.layer_sizes[-1]
        ), "sigma_x size is different output layer size"
        assert (
            self.mu_x.shape == self.sigma_x.shape
        ), "mu_x and sigma_x have different shapes"
        assert (
            self.mu_y.shape == self.sigma_y.shape
        ), "mu_y and sigma_y have different shapes"
        m = self.layer_sizes[0]
        for i, n in enumerate(self.layer_sizes):
            assert (
                self.a[i] in ACTIVATIONS
            ), f"a[{i}] must be one of {list(ACTIVATIONS.keys())}"
            assert self.b[i].shape == (
                n,
                1,
            ), f"b[{i}] has the wrong shape (expected {(n, 1)})"
            assert self.W[i].shape == (
                n,
                m,
            ), f"W[{i}] has the wrong shape (expected {(n, m)})"
            m = n

    def save(self, binary_file: Union[str, Path] = "parameters.json") -> None:
        """Save parameters to specified json file."""
        with open(binary_file, "wb") as file:
            file.write(self._serialize())

    def load(self, binary_file: Union[str, Path] = "parameters.json") -> None:
        """Load parameters from specified json file."""
        with open(binary_file, "rb") as file:
            byte_stream = file.read()
        self._deserialize(byte_stream)



"""Test forward and backward propagation using 
neural nets for which the correct value of the 
parameters is known exactly."""
import numpy as np
import pytest 
from copy import deepcopy
from typing import List

import jenn


def _finite_difference(cost, params, step=1e-6):
    """Use finite difference to compute partials of cost function
    with respect to neural net parameters (backpropagation)."""
    grads = list()
    dx = step
    for x in params:
        n, p = x.shape
        dy = np.zeros((n, p))
        for i in range(0, n):
            for j in range(0, p):
                # Forward step
                x[i, j] += dx
                y_fwd = cost(params)
                x[i, j] -= dx

                # Backward step
                x[i, j] -= dx
                y_bwd = cost(params)
                x[i, j] += dx

                # Central difference
                dy[i, j] = np.divide(y_fwd - y_bwd, 2 * dx)

        grads.append(dy)
    return grads


def _grad_check(dydx: List[np.ndarray], dydx_FD: List[np.ndarray],
                atol: float = 1e-6, rtol: float = 1e-4) -> bool:
    """
    Compare analytical gradient against finite difference

    Parameters
    ----------
    x: List[np.ndarray]
        Point at which to evaluate gradient

    f: callable
        Function handle to use for finite difference

    dx: float
        Finite difference step

    tol: float
        Tolerance below which agreement is considered acceptable
        Default = 1e-6

    verbose: bool
        Print output to standard out
        Default = True

    Returns
    -------
    success: bool
        Returns True iff finite difference and analytical grads agree
    """
    success = True
    for i in range(len(dydx)):
        if not np.allclose(dydx[i], dydx_FD[i], atol=atol, rtol=rtol):
            success = False
        if not success:
            msg = f"The gradients of layer {i} are wrong\n"
        else:
            msg = f"The gradients of layer {i} are correct\n"
        msg += f"Finite dif: grad[{i}] = {dydx_FD[i].squeeze()}\n"
        msg += f"Analytical: grad[{i}] = {dydx[i].squeeze()}\n\n"
        assert success, msg
    return success


X_test = np.array([[0, 1, 0, 1], [0, 0, 1, 1]])
Y_test = np.array([[0, 1, 1, 0]])


class TestXOR: 
    """Check forward and backprop on XOR test case."""

    @pytest.fixture
    def data(self) -> jenn.core.data.Dataset:
        """Return XOR test data."""
        return jenn.core.data.Dataset(X_test, Y_test)

    @pytest.fixture
    def cache(self) -> jenn.core.cache.Cache:
        """Return XOR cache."""
        return jenn.core.cache.Cache(layer_sizes=[2, 2, 1], m=Y_test.size)

    @pytest.fixture
    def params(self) -> jenn.core.parameters.Parameters:
        """Return XOR parameters."""
        parameters = jenn.core.parameters.Parameters(
            layer_sizes=[2, 2, 1], 
            hidden_activation='relu',
            output_activation='relu',
        )
        parameters.initialize()
        parameters.b[1][:] = np.array([[0], [-1]])        # layer 1
        parameters.W[1][:] = np.array([[1, 1], [1, 1]])   # layer 1
        parameters.b[2][:] = np.array([[0]])              # layer 2
        parameters.W[2][:] = np.array([[1, -2]])          # layer 2
        return parameters

    def test_model_forward(
            self, 
            data: jenn.core.data.Dataset, 
            params: jenn.core.parameters.Parameters, 
            cache: jenn.core.cache.Cache,
        ) -> None:
        """Test forward propagation using XOR."""
        computed = jenn.core.propagation.partials_forward(data.X, params, cache)
        expected = data.Y
        msg = f'computed = {computed} vs. expected = {expected}'
        assert np.all(computed == expected), msg


    def test_model_backward(
            self, 
            data: jenn.core.data.Dataset, 
            params: jenn.core.parameters.Parameters, 
            cache: jenn.core.cache.Cache,
        ) -> None:
        """Test backward propagation against finite difference."""

        ###########################
        # Perfectly trained model #
        ###########################

        jenn.core.propagation.model_partials_forward(
            data.X, params, cache)  # predict to populate cache

        jenn.core.propagation.model_backward(
            data, params, cache)  # partials computed in place

        dydx = params.stack_partials()
        assert np.allclose(dydx, 0.0)  # partials should be 0 at optimum params

        ###################
        # Imperfect model #
        ###################

        for i in range(params.L): # falsify model so partials are not zero
            params.W[i][:] += 10 * np.random.rand()
            params.b[i][:] += 10 * np.random.rand()

        jenn.core.propagation.model_partials_forward(
            data.X, params, cache)  # predict to populate cache

        jenn.core.propagation.model_backward(
            data, params, cache)  # partials computed in place

        def cost_FD(x):
            parameters = deepcopy(params)  # make copy b/c arrays updated in place
            cost = jenn.core.cost.Cost(data, parameters)
            parameters.unstack(x)
            Y_pred = jenn.core.propagation.model_forward(data.X, parameters, deepcopy(cache))
            return cost.evaluate(Y_pred)

        dydx = params.stack_partials_per_layer()
        dydx_FD = _finite_difference(cost_FD, params.stack_per_layer())

        assert _grad_check(dydx, dydx_FD)


# TODO: add test(s) for gradient-enhanced backprop and forward prop of partials


"""Model.
=========

This module contains the main class to train a neural net and make
predictions. It acts as an interface between the user and the core 
functions doing computations under-the-hood. 

.. code-block:: python

    #################
    # Example Usage #
    #################

    import jenn 

    # Fit model
    nn = jenn.model.NeuralNet(
        layer_sizes=[
            x_train.shape[0],  # input layer 
            7, 7,              # hidden layer(s) -- user defined
            y_train.shape[0]   # output layer 
         ],  
        ).fit(
            x_train, y_train, dydx_train, **kwargs # note: user must provide this
        )

    # Predict response only 
    y_pred = nn.predict(x_test)

    # Predict partials only 
    dydx_pred = nn.predict_partials(x_train)

    # Predict response and partials in one step (preferred)
    y_pred, dydx_pred = nn.evaluate(x_test) 

.. Note::
    The method `evaluate()` is preferred over separately 
    calling `predict()` followed by `predict_partials()` 
    whenever both the response and its partials are needed at the same point. 
    This saves computations since, in the latter approach, forward propagation 
    is unecessarily performed twice. Similarly, to avoid unecessary partial
    deerivative calculations, the `predict()` method should be preferred whenever
    only response values are needed. The method `predict_partials()` is provided 
    for those situations where it is necessary to separate out Jacobian predictions, 
    due to how some target optimization software architected for example.  
"""  # noqa: W291

from pathlib import Path
from typing import Any, List, Tuple, Union

import numpy as np

from .core.cache import Cache
from .core.data import Dataset, denormalize, denormalize_partials, normalize
from .core.parameters import Parameters
from .core.propagation import model_forward, model_partials_forward, partials_forward
from .core.training import train_model

__all__ = ["NeuralNet"]


class NeuralNet:
    """Neural network model.

    :param layer_sizes: number of nodes in each layer (including
        input/output layers)
    :param hidden_activation: activation function used in hidden layers
    :param output_activation: activation function used in output layer
    """

    def __init__(
        self,
        layer_sizes: List[int],
        hidden_activation: str = "tanh",
        output_activation: str = "linear",
    ):  # noqa D107
        self.history: Union[dict[Any, Any], None] = None
        self.parameters = Parameters(
            layer_sizes,
            hidden_activation,
            output_activation,
        )

    def fit(
        self,
        x: np.ndarray,
        y: np.ndarray,
        dydx: Union[np.ndarray, None] = None,
        is_normalize: bool = False,
        alpha: float = 0.05,
        beta: Union[np.ndarray, float] = 1.0,
        gamma: Union[np.ndarray, float] = 1.0,
        lambd: float = 0.0,
        beta1: float = 0.9,
        beta2: float = 0.99,
        tau: float = 0.5,
        tol: float = 1e-12,
        max_count: int = 1000,
        epsilon_absolute: float = 1e-12,
        epsilon_relative: float = 1e-12,
        epochs: int = 1,
        batch_size: Union[int, None] = None,
        max_iter: int = 1000,
        shuffle: bool = True,
        random_state: Union[int, None] = None,
        is_backtracking: bool = False,
        is_warmstart: bool = False,
        is_verbose: bool = False,
    ) -> "NeuralNet":  # noqa: PLR0913
        r"""Train neural network.

        :param x: training data inputs, array of shape (n_x, m)
        :param y: training data outputs, array of shape (n_y, m)
        :param dydx: training data Jacobian, array of shape (n_y, n_x, m)
        :param is_normalize: normalize training by mean and variance
        :param alpha: optimizer learning rate for line search
        :param beta: LSE coefficients [defaulted to one] (optional)
        :param gamma: jacobian-enhancement regularization coefficient [defaulted to zero] (optional)
        :param lambd: regularization coefficient to avoid overfitting [defaulted to zero] (optional)
        :param beta1: `ADAM <https://arxiv.org/abs/1412.6980>`_ optimizer hyperparameter to control momentum
        :param beta2: ADAM optimizer hyperparameter to control momentum
        :param tau: amount by which to reduce :math:`\alpha := \tau \times
            \alpha` on each iteration
        :param tol: stop when cost function doesn't improve more than
            specified tolerance
        :param max_count: stop when line search iterations exceed maximum
            count specified
        :param epsilon_absolute: absolute error stopping criterion
        :param epsilon_relative: relative error stopping criterion
        :param epochs: number of passes through data
        :param batch_size: size of each batch for minibatch
        :param max_iter: max number of optimizer iterations
        :param shuffle: shuffle minibatches or not
        :param random_state: control repeatability
        :param is_backtracking: use backtracking line search or not
        :param is_warmstart: do not initialize parameters
        :param is_verbose: print out progress for each (iteration, batch, epoch)
        :return: NeuralNet instance (self)

        .. warning::
                Normalization usually helps, except when the training
                data is made up of very small numbers. In that case,
                normalizing by the variance has the undesirable effect
                of dividing by a very small number and should not be used.
        """
        data = Dataset(x, y, dydx)
        params = self.parameters
        if not is_warmstart:
            params.initialize(random_state)
        params.mu_x[:] = 0.0
        params.mu_y[:] = 0.0
        params.sigma_x[:] = 1.0
        params.sigma_y[:] = 1.0
        if is_normalize:
            params.mu_x[:] = data.avg_x
            params.mu_y[:] = data.avg_y
            params.sigma_x[:] = data.std_x
            params.sigma_y[:] = data.std_y
            data = data.normalize()
        self.history = train_model(
            data,
            params,
            # hyperparameters
            alpha=alpha,
            beta=beta,
            gamma=gamma,
            lambd=lambd,
            beta1=beta1,
            beta2=beta2,
            tau=tau,
            tol=tol,
            max_count=max_count,
            epsilon_absolute=epsilon_absolute,
            epsilon_relative=epsilon_relative,
            # options
            epochs=epochs,
            max_iter=max_iter,
            batch_size=batch_size,
            shuffle=shuffle,
            random_state=random_state,
            is_backtracking=is_backtracking,
            is_verbose=is_verbose,
        )
        return self

    def predict(self, x: np.ndarray) -> np.ndarray:
        r"""Predict responses.

        :param x: vectorized inputs, array of shape (n_x, m)
        :return: predicted response(s), array of shape (n_y, m)
        """
        params = self.parameters
        cache = Cache(params.layer_sizes, m=x.shape[1])
        x_norm = normalize(x, params.mu_x, params.sigma_x)
        y_norm = model_forward(x_norm, params, cache)
        y = denormalize(y_norm, params.mu_y, params.sigma_y)
        return y

    def predict_partials(self, x: np.ndarray) -> np.ndarray:
        r"""Predict partials.

        :param x: vectorized inputs, array of shape (n_x, m)
        :return: predicted partial(s), array of shape (n_y, n_x, m)
        """
        params = self.parameters
        cache = Cache(params.layer_sizes, m=x.shape[1])
        x_norm = normalize(x, params.mu_x, params.sigma_x)
        dydx_norm = partials_forward(x_norm, params, cache)
        dydx = denormalize_partials(dydx_norm, params.sigma_x, params.sigma_y)
        return dydx

    def evaluate(self, x: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
        r"""Predict responses and their partials.

        :param x: vectorized inputs, array of shape (n_x, m)
        :return: predicted response(s), array of shape (n_y, m)
        :return: predicted partial(s), array of shape (n_y, n_x, m)
        """
        params = self.parameters
        cache = Cache(params.layer_sizes, m=x.shape[1])
        x_norm = normalize(x, params.mu_x, params.sigma_x)
        y_norm, dydx_norm = model_partials_forward(x_norm, params, cache)
        y = denormalize(y_norm, params.mu_y, params.sigma_y)
        dydx = denormalize_partials(dydx_norm, params.sigma_x, params.sigma_y)
        return y, dydx

    def save(self, file: Union[str, Path] = "parameters.json") -> None:
        """Serialize parameters and save to JSON file."""
        self.parameters.save(file)

    def load(self, file: Union[str, Path] = "parameters.json") -> "NeuralNet":
        """Load previously saved parameters from json file."""
        self.parameters.load(file)
        return self



"""Data.
========

This module contains convenience utilities to 
manage and handle training data. 
"""  # noqa: W291

import math
from dataclasses import dataclass
from functools import cached_property
from typing import List, Tuple, Union

import numpy as np


def mini_batches(
    X: np.ndarray,
    batch_size: Union[int, None],
    shuffle: bool = True,
    random_state: Union[int, None] = None,
) -> List[Tuple[int, ...]]:
    r"""Create randomized mini-batches.

    :param X: training data input :math:`X\in\mathbb{R}^{n_x\times m}`
    :param batch_size: mini batch size (if None, single batch with all
        data)
    :param shuffle: swhether to huffle data points or not
    :param random_state: random seed (useful to make runs repeatable)
    :return: list of tuples containing training data indices allocated
        to each batch
    """
    rng = np.random.default_rng(random_state)

    batches = []
    m = X.shape[1]
    if not batch_size:
        batch_size = m
    batch_size = min(batch_size, m)

    # Step 1: Shuffle the indices
    indices: list[int] = list(rng.permutation(m)) if shuffle else np.arange(m).tolist()

    # Step 2: Partition (shuffled_X, shuffled_Y). Minus the end case.
    num_complete_minibatches = int(math.floor(m / batch_size))
    k = 0
    for _ in range(num_complete_minibatches):
        mini_batch = indices[k * batch_size : (k + 1) * batch_size]
        if mini_batch:
            batches.append(tuple(mini_batch))
        k += 1

    # Handling the end case (last mini-batch < mini_batch_size)
    if m % batch_size != 0:
        mini_batch = indices[(k + 1) * batch_size :]
        if mini_batch:
            batches.append(tuple(mini_batch))

    return batches


def avg(array: np.ndarray) -> np.ndarray:
    """Compute mean and reshape as column array.

    :param array: array of shape (-1, m)
    :return: column array corresponding to mean of each row
    """
    return np.mean(array, axis=1).reshape((-1, 1))


def std(array: np.ndarray) -> np.ndarray:
    """Compute standard deviation and reshape as column array.

    :param array: array of shape (-1, m)
    :return: column array corresponding to std dev of each row
    """
    return np.std(array, axis=1).reshape((-1, 1))


def _safe_divide(
    value: np.ndarray, eps: float = float(np.finfo(float).eps)
) -> np.ndarray:
    """Add small number to avoid dividing by zero."""
    mask = value == 0.0  # noqa: PLR2004
    value[mask] += eps
    return value


def normalize(data: np.ndarray, mu: np.ndarray, sigma: np.ndarray) -> np.ndarray:
    """Center data about mean and normalize by standard deviation.

    :param data: data to be normalized, array of shape (-1, m)
    :param mu: mean of the data, array of shape (-1, 1)
    :param sigma: std deviation of the data, array of shape (-1, 1)
    :return: normalized data, array of shape (-1, m)
    """
    return (data - mu) / _safe_divide(sigma)


def denormalize(data: np.ndarray, mu: np.ndarray, sigma: np.ndarray) -> np.ndarray:
    """Undo normalization.

    :param data: normalized data, array of shape (-1, m)
    :param mu: mean of the data, array of shape (-1, 1)
    :param sigma: std deviation of the data, array of shape (-1, 1)
    :return: denormalized data, array of shape (-1, m)
    """
    return sigma * data + mu


def normalize_partials(
    partials: Union[np.ndarray, None], sigma_x: np.ndarray, sigma_y: np.ndarray
) -> Union[np.ndarray, None]:
    r"""Normalize partials.

    :param partials: training data partials to be normalized
        :math:`J\in\mathbb{R}^{n_y\times n_x \times m}`
    :param sigma_x: std dev of training data factors :math:`\sigma_x`,
        array of shape (-1, 1)
    :param sigma_y: std dev of training data responses :math:`\sigma_y`,
        array of shape (-1, 1)
    :return: normalized partials, array of shape (n_y, n_x, m)
    """
    if partials is None:
        return partials
    n_y, n_x, _ = partials.shape
    sigma_x = sigma_x.T.reshape((1, n_x, 1))
    sigma_y = sigma_y.reshape((n_y, 1, 1))
    return partials * sigma_x / _safe_divide(sigma_y)


def denormalize_partials(
    partials: np.ndarray, sigma_x: np.ndarray, sigma_y: np.ndarray
) -> np.ndarray:
    r"""Undo normalization of partials.

    :param partials: normalized training data partials
        :math:`\bar{J}\in\mathbb{R}^{n_y\times n_x \times m}`
    :param sigma_x: std dev of training data factors :math:`\sigma_x`,
        array of shape (-1, 1)
    :param sigma_y: std dev of training data responses :math:`\sigma_y`,
        array of shape (-1, 1)
    :return: denormalized partials, array of shape (n_y, n_x, m)
    """
    n_y, n_x, _ = partials.shape
    sigma_x = sigma_x.T.reshape((1, n_x, 1))
    sigma_y = sigma_y.reshape((n_y, 1, 1))
    return partials * sigma_y / _safe_divide(sigma_x)


@dataclass
class Dataset:
    """Store training data and associated metadata for easy access.

    :param X: training data outputs, array of shape (n_x, m)
    :param Y: training data outputs, array of shape (n_y, m)
    :param J: training data Jacobians, array of shape (n_y, n_x, m)
    """

    X: np.ndarray
    Y: np.ndarray
    J: Union[np.ndarray, None] = None

    Y_weights: Union[np.ndarray, float] = 1.0
    J_weights: Union[np.ndarray, float] = 1.0

    def __post_init__(self) -> None:  # noqa: D105
        if self.X.shape[1] != self.Y.shape[1]:
            msg = "X and Y must have the same number of examples"
            raise ValueError(msg)

        n_y, n_x, m = self.n_y, self.n_x, self.m

        self.Y_weights = self.Y_weights * np.ones((n_y, m))
        self.J_weights = self.J_weights * np.ones((n_y, n_x, m))

        if self.J is not None:
            if self.J.shape != (n_y, n_x, m):
                msg = f"J must be of shape ({n_y}, {n_x}, {m})"
                raise ValueError(msg)

    def set_weights(
        self,
        beta: Union[np.ndarray, float] = 1.0,
        gamma: Union[np.ndarray, float] = 1.0,
    ) -> None:
        """Prioritize certain points more than others.

        Rational: this can be used to reward the optimizer more in certain regions.

        :param beta: multiplier(s) on Y
        :param beta: multiplier(s) on J
        """
        self.Y_weights = beta * np.ones((self.n_y, self.m))
        self.J_weights = gamma * np.ones((self.n_y, self.n_x, self.m))

    @property
    def m(self) -> int:
        """Return number of training examples."""
        return int(self.X.shape[1])

    @property
    def n_x(self) -> int:
        """Return number of inputs."""
        return int(self.X.shape[0])

    @property
    def n_y(self) -> int:
        """Return number of outputs."""
        return int(self.Y.shape[0])

    @cached_property
    def avg_x(self) -> np.ndarray:
        """Return mean of input data as array of shape (n_x, 1)."""
        return avg(self.X)

    @cached_property
    def avg_y(self) -> np.ndarray:
        """Return mean of output data as array of shape (n_y, 1)."""
        return avg(self.Y)

    @cached_property
    def std_x(self) -> np.ndarray:
        """Return standard dev of input data, array of shape (n_x, 1)."""
        return std(self.X)

    @cached_property
    def std_y(self) -> np.ndarray:
        """Return standard dev of output data, array of shape (n_y, 1)."""
        return std(self.Y)

    def mini_batches(
        self,
        batch_size: Union[int, None],
        shuffle: bool = True,
        random_state: Union[int, None] = None,
    ) -> List["Dataset"]:
        """Breakup data into multiple batches and return list of Datasets.

        :param batch_size: mini batch size (if None, single batch with
            all data)
        :param shuffle: swhether to huffle data points or not
        :param random_state: random seed (useful to make runs
            repeatable)
        :return: list of Dataset representing data broken up in batches
        """
        X = self.X
        Y = self.Y
        J = self.J
        Y_weights = np.ones(Y.shape) * self.Y_weights
        batches = mini_batches(X, batch_size, shuffle, random_state)
        if J is None:
            return [
                Dataset(X[:, b], Y[:, b], Y_weights=Y_weights[:, b]) for b in batches
            ]
        J_weights = np.ones(J.shape) * self.J_weights
        return [
            Dataset(X[:, b], Y[:, b], J[:, :, b], Y_weights[:, b], J_weights[:, :, b])
            for b in batches
        ]

    def normalize(self) -> "Dataset":
        """Return normalized Dataset."""
        X_norm = normalize(self.X, self.avg_x, self.std_x)
        Y_norm = normalize(self.Y, self.avg_y, self.std_y)
        J_norm = normalize_partials(self.J, self.std_x, self.std_y)
        return Dataset(X_norm, Y_norm, J_norm)



"""Training.
============

This class implements the core algorithm responsible for training the neural networks."""

import functools
from collections import defaultdict
from typing import Union

import numpy as np

from .cache import Cache
from .cost import Cost
from .data import Dataset
from .optimization import ADAMOptimizer
from .parameters import Parameters
from .propagation import model_backward, model_partials_forward


def objective_function(
    X: np.ndarray,
    cost: Cost,
    parameters: Parameters,
    cache: Cache,
    stacked_params: np.ndarray,
) -> np.float64:
    """Evaluate cost function for training.

    :param X: training data inputs, array of shape (n_x, m)
    cost: cost function to be evaluated
    :param parameters: object that stores neural net parameters for each
        layer
    :param cache: neural net cache that stores neural net quantities
        computed during forward prop for each layer, so they can be
        accessed during backprop to avoid re-computing them
    stacked_params: neural network parameters returned by the optimizer,
        represented as single array of stacked parameters for all layers.
    """
    parameters.unstack(stacked_params)
    Y_pred, J_pred = model_partials_forward(X, parameters, cache)
    return cost.evaluate(Y_pred, J_pred)


def objective_gradient(
    data: Dataset,
    parameters: Parameters,
    cache: Cache,
    lambd: float,
    stacked_params: np.ndarray,
) -> np.ndarray:  # noqa: PLR0913
    """Evaluate cost function gradient for backprop.

    :param data: object containing training and associated metadata
    :param parameters: object that stores neural net parameters for each
        layer
    :param cache: neural net cache that stores neural net quantities
        computed during forward prop for each layer, so they can be
        accessed during backprop to avoid re-computing them
    :param lambd: coefficient that multiplies regularization term in
        cost function
    :param gamma: coefficient that multiplies jacobian-enhancement term
        in cost function
    :param stacked_params: neural network parameters returned by the
        optimizer, represented as single array of stacked parameters for
        all layers.
    """
    parameters.unstack(stacked_params)
    model_backward(data, parameters, cache, lambd)
    return parameters.stack_partials()


def train_model(
    data: Dataset,
    parameters: Parameters,
    alpha: float = 0.05,
    beta: Union[np.ndarray, float] = 1.0,
    gamma: Union[np.ndarray, float] = 1.0,
    lambd: float = 0.0,
    beta1: float = 0.9,
    beta2: float = 0.99,
    tau: float = 0.5,
    tol: float = 1e-12,
    max_count: int = 1000,
    epsilon_absolute: float = 1e-12,
    epsilon_relative: float = 1e-12,
    epochs: int = 1,
    max_iter: int = 200,
    batch_size: Union[int, None] = None,
    shuffle: bool = True,
    random_state: Union[int, None] = None,
    is_backtracking: bool = False,
    is_verbose: bool = False,
) -> dict:  # noqa: PLR0913
    r"""Train neural net.

    :param data: object containing training and associated metadata
    :param parameters: object that stores neural net parameters for each
        layer
    :param alpha: learning rate :math:`\alpha`
    :param beta: LSE coefficients [defaulted to one] (optional)
    :param gamma: jacobian-enhancement regularization coefficient
        [defaulted to zero] (optional)
    :param lambd: regularization coefficient to avoid overfitting
        [defaulted to zero] (optional)
    :param beta_1: exponential decay rate of 1st moment vector
        :math:`\beta_1\in[0, 1)`
    :param beta_2: exponential decay rate of 2nd moment vector
        :math:`\beta_2\in[0, 1)`
    :param tau: amount by which to reduce :math:`\alpha := \tau \times
        \alpha` on each iteration
    :param tol: stop when cost function doesn't improve more than
        specified tolerance
    :param max_count: stop when line search iterations exceed maximum
        count specified
    :param epsilon_absolute: absolute error stopping criterion
    :param epsilon_relative: relative error stopping criterion
    :param epochs: number of passes through data
    :param batch_size: mini batch size (if None, single batch with all
        data)
    :param max_iter: maximum number of optimizer iterations allowed
    :param shuffle: swhether to huffle data points or not
    :param random_state: random seed (useful to make runs repeatable)
    :param is_backtracking: whether or not to use backtracking during
        line search
    :param is_verbose: print out progress for each iteration, each
        batch, each epoch
    :return: cost function training history accessed as `cost =
        history[epoch][batch][iter]`
    """
    history: dict[str, dict[str, Union[list[np.ndarray], None]]] = defaultdict(dict)
    optimizer = ADAMOptimizer(
        beta_1=beta1,
        beta_2=beta2,
        tau=tau,
        tol=tol,
        max_count=is_backtracking * max_count,
    )
    data.set_weights(beta, gamma)
    for e in range(epochs):
        batches = data.mini_batches(batch_size, shuffle, random_state)
        for b, batch in enumerate(batches):
            cache = Cache(parameters.layer_sizes, batch.m)
            cost = Cost(batch, parameters, lambd)
            func = functools.partial(
                objective_function,
                batch.X,
                cost,
                parameters,
                cache,
            )
            grad = functools.partial(
                objective_gradient,
                batch,
                parameters,
                cache,
                lambd,
            )
            optimizer.minimize(
                x=parameters.stack(),
                f=func,
                dfdx=grad,
                alpha=alpha,
                max_iter=max_iter,
                epsilon_absolute=epsilon_absolute,
                epsilon_relative=epsilon_relative,
                verbose=is_verbose,
                epoch=e,
                batch=b,
            )
            history[f"epoch_{e}"][f"batch_{b}"] = optimizer.cost_history
    return history



"""Utils module entry point."""

# Copyright (c) 2018 Steven H. Berguin
# Distributed under the terms of the MIT License.

from . import metrics, plot
from .jmp import from_jmp
from .rbf import rbf

__all__ = [
    "metrics",
    "plot",
    "rbf",
    "from_jmp",
]



"""Metrics.
===========
"""  # noqa: W291

import numpy as np


def r_square(y_pred: np.ndarray, y_true: np.ndarray) -> np.ndarray:
    """Compute R-square value for each output.

    :param y_pred: predicted values, array of shape (n_y, m)
    :param y_true: actuial values, array of shape (n_y, m)
    :return: R-Squared values for each predicted reponse
    """
    axis = y_true.ndim - 1
    y_bar = np.mean(y_true, axis=axis)
    SSE = np.sum(np.square(y_pred - y_true), axis=axis)
    SSTO = np.sum(np.square(y_true - y_bar) + 1e-12, axis=axis)
    return 1 - SSE / SSTO



"""Activation.
==============

This module implements activation functions used by the neural network."""

import abc
from typing import Union

import numpy as np


class Activation:
    """Activation function base class."""

    @classmethod
    @abc.abstractmethod
    def evaluate(
        cls,
        x: np.ndarray,
        y: Union[np.ndarray, None] = None,
    ) -> np.ndarray:  # noqa: D102
        """Evaluate activation function.

        :param x: input array at which to evaluate the function
        :param y: output array in which to write the results (optional)
        :return: activation function evaluated at `x` (as new array if `y` not provided as input)
        """
        raise NotImplementedError

    @classmethod
    @abc.abstractmethod
    def first_derivative(
        cls,
        x: np.ndarray,
        y: Union[np.ndarray, None] = None,
        dy: Union[np.ndarray, None] = None,
    ) -> np.ndarray:  # noqa: D102
        """Evaluate 1st derivative of activation function.

        :param x: input array at which to evaluate the function
        :param y: response already evaluated at x (optional)
        :param dy: output array in which to write the 1st derivative (optional)
        :return: 1st derivative (as new array if `dy` not provided as input)
        """
        raise NotImplementedError

    @classmethod
    @abc.abstractmethod
    def second_derivative(
        cls,
        x: np.ndarray,
        y: Union[np.ndarray, None] = None,
        dy: Union[np.ndarray, None] = None,
        ddy: Union[np.ndarray, None] = None,
    ) -> np.ndarray:  # noqa: D102
        """Evaluate 2nd derivative of activation function.

        :param x: input array at which to evaluate the function
        :param y: response already evaluated at x (optional)
        :param y: 1st derivative already evaluated at x (optional)
        :param ddy: output array in which to write the 2nd derivative (optional)
        :return: 2nd derivative (as new array if `ddy` not provided as input)
        """
        raise NotImplementedError


class Tanh(Activation):
    r"""Hyperbolic tangent.

    .. math::
        y = \tanh(x) = \frac{e^x - e^{-x}}{e^x + e^{-x}}
    """

    @classmethod
    def evaluate(
        cls,
        x: np.ndarray,
        y: Union[np.ndarray, None] = None,
    ) -> np.ndarray:  # noqa: D102
        return np.tanh(x, out=y)  # evaluated in place if y is not None

    @classmethod
    def first_derivative(
        cls,
        x: np.ndarray,
        y: Union[np.ndarray, None] = None,
        dy: Union[np.ndarray, None] = None,
    ) -> np.ndarray:  # noqa: D102
        if y is None:
            y = cls.evaluate(x)
        if dy is None:
            return 1 - np.square(y)
        dy[:] = 1 - np.square(y, out=dy)
        return dy

    @classmethod
    def second_derivative(
        cls,
        x: np.ndarray,
        y: Union[np.ndarray, None] = None,
        dy: Union[np.ndarray, None] = None,
        ddy: Union[np.ndarray, None] = None,
    ) -> np.ndarray:  # noqa: D102
        if y is None:
            y = cls.evaluate(x)
        if dy is None:
            dy = cls.first_derivative(x, y)
        if ddy is None:
            return -2 * y * dy
        ddy[:] = -2 * y * dy
        return ddy


class Relu(Activation):
    r"""Rectified linear unit activation.

    .. math::
        y = \begin{cases}
            x & \text{if}~ x \ge 0 \\
            0 & \text{otherwise}
        \end{cases}
    """

    @classmethod
    def evaluate(
        cls,
        x: np.ndarray,
        y: Union[np.ndarray, None] = None,
    ) -> np.ndarray:  # noqa: D102
        if y is None:
            y = (x > 0) * x
        else:
            y[:] = (x > 0) * x
        return y

    @classmethod
    def first_derivative(
        cls,
        x: np.ndarray,
        y: Union[np.ndarray, None] = None,
        dy: Union[np.ndarray, None] = None,
    ) -> np.ndarray:  # noqa: D102
        if dy is None:
            dy = np.asarray(x > 0, dtype=x.dtype)
        else:
            dy[:] = np.asarray(x > 0, dtype=x.dtype)
        return dy

    @classmethod
    def second_derivative(
        cls,
        x: np.ndarray,
        y: Union[np.ndarray, None] = None,
        dy: Union[np.ndarray, None] = None,
        ddy: Union[np.ndarray, None] = None,
    ) -> np.ndarray:  # noqa: D102
        if ddy is None:
            return np.zeros(x.shape)
        ddy[:] = 0.0
        return ddy


class Linear(Activation):
    r"""Linear activation function.

    .. math::
        y = x
    """

    @classmethod
    def evaluate(
        cls,
        x: np.ndarray,
        y: Union[np.ndarray, None] = None,
    ) -> np.ndarray:  # noqa: D102
        if y is None:
            y = x.copy()
        else:
            y[:] = x
        return y

    @classmethod
    def first_derivative(
        cls,
        x: np.ndarray,
        y: Union[np.ndarray, None] = None,
        dy: Union[np.ndarray, None] = None,
    ) -> np.ndarray:  # noqa: D102
        if dy is None:
            dy = np.ones(x.shape)
        else:
            dy[:] = 1
        return dy

    @classmethod
    def second_derivative(
        cls,
        x: np.ndarray,
        y: Union[np.ndarray, None] = None,
        dy: Union[np.ndarray, None] = None,
        ddy: Union[np.ndarray, None] = None,
    ) -> np.ndarray:  # noqa: D102
        if ddy is None:
            return np.zeros(x.shape)
        ddy[:] = 0
        return ddy


ACTIVATIONS = dict(
    relu=Relu,
    tanh=Tanh,
    linear=Linear,
)



"""Test Parameter class."""
import pytest 
import numpy as np 
import tempfile
import pathlib

import jenn 


class TestSerialization: 
    """Check that parameters can be saved and reloaded."""

    @pytest.fixture
    def params(self) -> jenn.core.parameters.Parameters:
        """Return XOR parameters."""
        parameters = jenn.core.parameters.Parameters(layer_sizes=[2, 2, 1], output_activation='relu')
        parameters.initialize()
        parameters.b[1][:] = np.array([[0], [-1]])        # layer 1
        parameters.W[1][:] = np.array([[1, 1], [1, 1]])   # layer 1
        parameters.b[2][:] = np.array([[0]])              # layer 2
        parameters.W[2][:] = np.array([[1, -2]])          # layer 2
        return parameters
    
    def test_serialization(self, params: jenn.core.parameters.Parameters) -> None: 
        """Test that saved parameters can be reloaded into a new object."""
        with tempfile.TemporaryDirectory() as tmpdirname:
            tmpfile = pathlib.Path(tmpdirname) / 'params.json'
            params.save(tmpfile)
            parameters = jenn.core.parameters.Parameters(params.layer_sizes) 
            assert params != parameters 
            parameters.load(tmpfile)
            assert params == parameters


"""Synthetic Data.
==================

This module provide synthetic test functions 
that can be used to generate exmaple data for 
illustration and testing. Simply inherit from 
the base class to implement new test functions. 

.. code-block:: python

    #################
    # Example Usage #
    #################

    import jenn 

    (
        x_train, 
        y_train, 
        dydx_train,
    ) = jenn.synthetic.Sinusoid.sample(
        m_lhs=0,    # number latin hypercube samples 
        m_levels=4, # number of full factorial levels per factor
        lb=-3.14,   # lower bound of domain 
        ub=3.14,    # upper bound of domain 
    )

    (
        x_test, 
        y_test, 
        dydx_test,
    ) = jenn.synthetic.Sinusoid.sample(
        m_lhs=30, 
        m_levels=0, 
        lb=-3.14,
        ub=3.14,
    )
"""  # noqa: W291

import abc
from typing import Tuple, Union

import numpy as np


def _fullfact(n_x: int, m_levels: int) -> np.ndarray:
    """Return full factorial with sample values between 0 and 1."""
    array = np.linspace(0, 1, m_levels)
    arrays = [array] * n_x
    meshes = [mesh.reshape((1, -1)) for mesh in np.meshgrid(*arrays)]
    return np.concatenate(meshes)


class TestFunction:
    """Test function base class."""

    @abc.abstractmethod
    def evaluate(self, x: np.ndarray) -> np.ndarray:
        """Evaluate function.

        :param x: inputs, array of shape (n_x, m)
        :return: response, array of shape (n_y, m)
        """
        raise NotImplementedError

    @abc.abstractmethod
    def first_derivative(self, x: np.ndarray) -> np.ndarray:
        """Evaluate partial derivative.

        :param x: inputs, array of shape (n_x, m)
        :return: partials, array of shape (n_y, n_x, m)
        """
        raise NotImplementedError

    @classmethod
    def first_derivative_FD(
        cls,
        x: np.ndarray,
        dx: float = 1e-6,
    ) -> np.ndarray:
        """Evaluate partial derivative using finite difference.

        :param x: inputs, array of shape (n_x, m)
        :return: partials, array of shape (n_y, n_x, m)
        """
        f = cls.evaluate
        y = f(x)  # type: ignore
        n_x, m = x.shape
        n_y = y.shape[0]
        dydx = np.zeros((n_y, n_x, m))
        for i in range(n_x):
            dx1 = np.zeros((n_x, m))
            dx2 = np.zeros((n_x, m))
            dx1[i] += dx
            dx2[i] += dx
            x1 = x - dx1
            x2 = x + dx2
            y1 = f(x1)  # type: ignore
            y2 = f(x2)  # type: ignore
            dydx[:, i] = (y2 - y1) / (2 * dx)
        return dydx

    @classmethod
    def sample(
        cls,
        m_lhs: int,
        m_levels: int,
        lb: Union[np.ndarray, float],
        ub: Union[np.ndarray, float],
        dx: Union[float, None] = 1e-6,
        random_state: Union[int, None] = None,
    ) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
        """Generate synthetic data by sampling the test function.

        :param m_lhs: number of latin hypercube samples
        :param m_levels: number of levels per factor for full factorial
        :param lb: lower bound on the factors
        :param ub: upper bound on the factors
        :param dx: finite difference step size (if None, analytical
            partials are used)
        :param random_state: random seed (for repeatability)
        """
        rng = np.random.default_rng(seed=random_state)
        lb = np.array([lb]).reshape((-1, 1))  # make sure it's an numpy array
        ub = np.array([ub]).reshape((-1, 1))  # make sure it's an numpy array
        n_x = lb.size
        lh = rng.random(size=(n_x, m_lhs))
        ff = _fullfact(n_x, m_levels)
        doe = np.concatenate([lh, ff], axis=1)
        m = doe.shape[1]
        x = lb + (ub - lb) * doe
        y = cls.evaluate(x).reshape((-1, m))  # type: ignore[call-arg]
        if dx is None:
            dydx = cls.first_derivative(x).reshape((-1, n_x, m))  # type: ignore[call-arg]
        else:
            dydx = cls.first_derivative_FD(x, dx).reshape((-1, n_x, m))
        return x, y, dydx


class Linear(TestFunction):
    r"""Linear function.

    .. math::
        f(x) = \beta_0 + \sum_{i=1}^p \beta_i x_i
    """

    @classmethod
    def evaluate(
        cls,
        x: np.ndarray,
        a: Union[float, np.ndarray] = 1.0,
        b: float = 0.0,
    ) -> np.ndarray:  # noqa: D102
        n_y = 1
        n_x, m = x.shape
        y = np.zeros((n_y, m))
        y[:] = a * np.sum(x, axis=0) + b
        return y

    @classmethod
    def first_derivative(
        cls,
        x: np.ndarray,
        a: Union[float, np.ndarray] = 1.0,
        b: float = 0.0,
    ) -> np.ndarray:  # noqa: D102
        n_y = 1
        n_x, m = x.shape
        dydx = np.zeros((n_y, n_x, m))
        a = np.array([a] * n_x) if isinstance(a, float) else a
        for i in range(n_x):
            dydx[0, i, :] = a[i]
        return dydx

    @classmethod
    def sample(
        cls,
        m_lhs: int = 100,
        m_levels: int = 0,
        lb: Union[np.ndarray, float] = -1.0,
        ub: Union[np.ndarray, float] = 1.0,
        dx: Union[float, None] = 1e-6,
        random_state: Union[int, None] = None,
    ) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:  # noqa: D102
        return super().sample(m_lhs, m_levels, lb, ub, dx, random_state)


class Parabola(TestFunction):
    r"""Parabolic function.

    .. math::
        f(x) = \frac{1}{n} \sum_{i=1}^p (x_i - {x_0}_i)^2
    """

    @classmethod
    def evaluate(
        cls, x: np.ndarray, x0: Union[np.ndarray, float] = 0.0
    ) -> np.ndarray:  # noqa: D102
        n_y = 1
        n_x, m = x.shape
        y = np.zeros((n_y, m))
        y[:] = 1 / n_x * np.sum((x - x0) ** 2, axis=0)
        return y

    @classmethod
    def first_derivative(
        cls, x: np.ndarray, x0: Union[np.ndarray, float] = 0.0
    ) -> np.ndarray:  # noqa: D102
        n_y = 1
        n_x, m = x.shape
        dydx = np.zeros((n_y, n_x, m))
        dydx[0, :, :] = 2 / n_x * (x - x0)
        return dydx

    @classmethod
    def sample(
        cls,
        m_lhs: int = 100,
        m_levels: int = 0,
        lb: Union[np.ndarray, float] = -1.0,
        ub: Union[np.ndarray, float] = 1.0,
        dx: Union[float, None] = 1e-6,
        random_state: Union[int, None] = None,
    ) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:  # noqa: D102
        return super().sample(m_lhs, m_levels, lb, ub, dx, random_state)


class Sinusoid(TestFunction):
    r"""Sinusoidal function.

    .. math::
        f(x) = x \sin(x)
    """

    @classmethod
    def evaluate(cls, x: np.ndarray) -> np.ndarray:  # noqa: D102
        n_y = 1
        n_x, m = x.shape
        y = np.zeros((n_y, m))
        y[:] = x * np.sin(x)
        return y

    @classmethod
    def first_derivative(cls, x: np.ndarray) -> np.ndarray:  # noqa: D102
        n_y = 1
        n_x, m = x.shape
        dydx = np.zeros((n_y, n_x, m))
        dydx[0, 0, :] = np.sin(x) + x * np.cos(x)
        return dydx

    @classmethod
    def sample(
        cls,
        m_lhs: int = 100,
        m_levels: int = 0,
        lb: Union[np.ndarray, float] = -np.pi,
        ub: Union[np.ndarray, float] = np.pi,
        dx: Union[float, None] = 1e-6,
        random_state: Union[int, None] = None,
    ) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:  # noqa: D102
        return super().sample(m_lhs, m_levels, lb, ub, dx, random_state)


class Rastrigin(TestFunction):
    r"""Rastrigin function.

    .. math::
        f(x) = \sum_{i=1}^p ( x_i^2 - 10 \cos(2\pi x_i) )
    """

    @classmethod
    def evaluate(cls, x: np.ndarray) -> np.ndarray:  # noqa: D102
        n_y = 1
        n_x, m = x.shape
        y = np.zeros((n_y, m)) + 10 * n_x
        for i in range(n_x):
            y += np.power(x[i], 2) - 10 * np.cos(2 * np.pi * x[i])
        return y

    @classmethod
    def first_derivative(cls, x: np.ndarray) -> np.ndarray:  # noqa: D102
        n_y = 1
        n_x, m = x.shape
        dydx = np.zeros((n_y, n_x, m))
        for i in range(n_x):
            dydx[0, i, :] = 2 * x[i] + 20 * np.pi * np.sin(2 * np.pi * x[i])
        return dydx

    @classmethod
    def sample(
        cls,
        m_lhs: int = 100,
        m_levels: int = 0,
        lb: Union[np.ndarray, float] = -1.0
        * np.ones(
            2,
        ),
        ub: Union[np.ndarray, float] = 1.5
        * np.ones(
            2,
        ),
        dx: Union[float, None] = 1e-6,
        random_state: Union[int, None] = None,
    ) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:  # noqa: D102
        return super().sample(m_lhs, m_levels, lb, ub, dx, random_state)


class Rosenbrock(TestFunction):
    r"""Banana Rosenbrock function.

    .. math::
        f(x) = (1 - x_1)^2 + 100 (x_2 - x_1^2)^ 2
    """

    @classmethod
    def evaluate(cls, x: np.ndarray) -> np.ndarray:  # noqa: D102
        n_y = 1
        n_x, m = x.shape
        y = np.zeros((n_y, m))
        y[:] = (1 - x[0]) ** 2 + 100 * (x[1] - x[0] ** 2) ** 2
        return y

    @classmethod
    def first_derivative(cls, x: np.ndarray) -> np.ndarray:  # noqa: D102
        n_y = 1
        n_x, m = x.shape
        dydx = np.zeros((n_y, n_x, m))
        dydx[0, 0, :] = -2 * (1 - x[0]) - 400 * x[0] * (x[1] - x[0] ** 2)
        dydx[0, 1, :] = 200 * (x[1] - x[0] ** 2)
        return dydx

    @classmethod
    def sample(
        cls,
        m_lhs: int = 100,
        m_levels: int = 0,
        lb: Union[np.ndarray, float] = -2
        * np.ones(
            2,
        ),
        ub: Union[np.ndarray, float] = 2.0
        * np.ones(
            2,
        ),
        dx: Union[float, None] = 1e-6,
        random_state: Union[int, None] = None,
    ) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:  # noqa: D102
        return super().sample(m_lhs, m_levels, lb, ub, dx, random_state)



"""Plotting.
============

This module provides optional but helpful utilities to 
assess goodness of fit and visualize trends. 

.. code-block:: python

    #################
    # Example Usage #
    #################

    import jenn 

    # Assuming the following are available: 
    x_train, y_train, dydx_train = _ # user provided
    x_test, y_test, dydx_test = _    # user provided
    nn = _                           # previously trained NeuralNet

    # Show goodness of fit of the partials 
    i = 0  # index of the response to plot
    jenn.utils.plot.goodness_of_fit(
        y_true=dydx_test[i], 
        y_pred=nn.predict_partials(x_test)[i], 
        title="Partial Derivative: dy/dx (NN)"
    )

    # Example: visualize local trends
    jenn.utils.plot.sensitivity_profiles(
        f=[nn.predict], 
        x_min=x_train.min(), 
        x_max=x_train.max(), 
        x_true=x_train, 
        y_true=y_train, 
        resolution=100, 
        legend=['nn'], 
        xlabels=['x'], 
        ylabels=['y'],
    )
"""  # noqa: W291

from collections.abc import Callable
from functools import wraps
from importlib.util import find_spec
from typing import Any, Dict, List, Tuple, Union

import numpy as np

from .metrics import r_square

if find_spec("matplotlib"):
    MATPLOTLIB_INSTALLED = True
    import matplotlib.pyplot as plt
else:
    MATPLOTLIB_INSTALLED = False


LINE_STYLES = {
    "solid": "solid",  # Same as (0, ()) or '-'
    "dotted": "dotted",  # Same as (0, (1, 1)) or ':'
    "dashdot": "dashdot",  # Same as '-.'
    "dashed": "dashed",  # Same as '--'
    "loosely dotted": (0, (1, 10)),
    # "dotted": (0, (1, 1)),
    "densely dotted": (0, (1, 1)),
    "long dash with offset": (5, (10, 3)),
    "loosely dashed": (0, (5, 10)),
    # "dashed": (0, (5, 5)),
    "densely dashed": (0, (5, 1)),
    "loosely dashdotted": (0, (3, 10, 1, 10)),
    "dashdotted": (0, (3, 5, 1, 5)),
    "densely dashdotted": (0, (3, 1, 1, 1)),
    "dashdotdotted": (0, (3, 5, 1, 5, 1, 5)),
    "loosely dashdotdotted": (0, (3, 10, 1, 10, 1, 10)),
    "densely dashdotdotted": (0, (3, 1, 1, 1, 1, 1)),
}


def requires_matplotlib(func: Callable) -> Callable:
    """Return error if matplotlib not installed."""

    @wraps(func)
    def wrapper(*args: list, **kwargs: dict) -> Any:  # noqa: ANN401
        if MATPLOTLIB_INSTALLED:
            return func(*args, **kwargs)
        raise ValueError("Matplotlib is not installed.")

    return wrapper


@requires_matplotlib
def actual_by_predicted(
    y_pred: np.ndarray,
    y_true: np.ndarray,
    ax: Union[plt.Axes, None] = None,  # noqa: ANN401
    figsize: Tuple[float, float] = (3.25, 3),
    title: str = "",
    fontsize: int = 9,
    alpha: float = 1.0,
) -> plt.Figure:  # noqa: ANN401
    """Create actual by predicted plot for a single response.

    :param y_pred: predicted values, array of shape (m,)
    :param y_true: true values, array of shape (m,)
    :param ax: the matplotlib axes on which to plot the data
    :param figsize: figure size
    :param title: title of figure
    :param fontsize: text size
    :param alpha: transparency of dots (between 0 and 1)
    :return: matplotlib figure instance
    """
    if y_pred.ndim > 1:
        if y_pred.ndim == 2 and 1 in y_pred.shape:
            pass  # just needs to be unraveled
        else:
            raise ValueError(
                f"Expected one dimensional array, "
                f"but y_pred has {y_pred.ndim} dimensions."
            )
    if y_true.ndim > 1 and 1 not in y_true.shape:
        if y_true.ndim == 2 and 1 in y_true.shape:
            pass  # just needs to be unraveled
        else:
            raise ValueError(
                f"Expected one dimensional array, "
                f"but y_true has {y_true.ndim} dimensions."
            )
    fig = plt.figure(figsize=figsize, layout="tight")
    if not ax:
        spec = fig.add_gridspec(ncols=1, nrows=1)
        ax = fig.add_subplot(spec[0, 0])
    actual = y_true.ravel()
    predicted = y_pred.ravel()
    ax.scatter(actual, predicted, color="k", alpha=alpha)
    line = [actual.min(), actual.max()]
    ax.plot(line, line, color="r", linestyle=":")
    ax.set_xlabel("Predicted", fontsize=fontsize)
    ax.set_ylabel("Actual", fontsize=fontsize)
    ax.set_title(title, fontsize=fontsize)
    ax.grid(True)
    ax.legend(["predictions", "perfect fit line"], fontsize=fontsize)
    plt.close(fig)
    return fig


def contours(
    func: Callable,
    lb: Tuple[float, float],
    ub: Tuple[float, float],
    x_train: Union[np.ndarray, None] = None,
    x_test: Union[np.ndarray, None] = None,
    figsize: Tuple[float, float] = (3.25, 3),
    fontsize: int = 9,
    alpha: float = 0.5,
    title: str = "",
    xlabel: str = "",
    ylabel: str = "",
    levels: int = 20,
    resolution: int = 100,
    ax: Union[plt.Axes, None] = None,  # noqa: ANN401
) -> Union[None, plt.Figure]:  # noqa: ANN401
    """Plot contours of a scalar function of two variables.

    :param figsize: figure size
    :param fontsize: text size
    :param alpha: transparency of dots (between 0 and 1)
    :param title: title of figure
    :param xlabel: factor #1 label
    :param ylabel: factor #2 label
    :param levels: number of contour levels
    :param resolution: line resolution
    :param ax: the matplotlib axes on which to plot the data
    :return: matplotlib figure instance
    """
    # Domain
    m = resolution
    x1 = np.linspace(lb[0], ub[0], m)
    x2 = np.linspace(lb[1], ub[1], m)
    x1, x2 = np.meshgrid(x1, x2)

    # Response
    y = np.zeros((m, m))
    for i in range(m):
        for j in range(m):
            y[i, j] = func(np.array([[x1[i, j]], [x2[i, j]]])).ravel()[0]

    # Plot
    if ax:
        fig = ax.get_figure()
    else:
        fig = plt.figure(figsize=figsize)
        ax = plt.gca()
    ax.contour(x1, x2, y, levels, cmap="RdGy", alpha=alpha)
    legend = []
    if x_train is not None:
        ax.scatter(x_train[0], x_train[1], marker=".", c="k", alpha=1)
        legend.append("train")
    if x_test is not None:
        ax.scatter(x_test[0], x_test[1], marker="+", c="r", alpha=1)
        legend.append("test")
    if legend:
        ax.legend(legend, loc=1)
    ax.set_title(title, fontsize=fontsize)
    ax.set_xlabel(xlabel, fontsize=fontsize)
    ax.set_ylabel(ylabel, fontsize=fontsize)
    plt.close(fig)
    return fig


@requires_matplotlib
def convergence(
    histories: List[Dict[str, Dict[str, List[float]]]],
    figsize: Tuple[float, float] = (3.25, 3),
    fontsize: int = 9,
    alpha: float = 1.0,
    title: str = "",
    legend: Union[List[str], None] = None,
) -> Union[plt.Figure, None]:  # noqa: ANN401
    """Plot training history.

    :param histories: training history for each model
    :param figsize: subfigure size of each subplot
    :param fontsize: text size
    :param alpha: transparency of dots (between 0 and 1)
    :param title: title of figure
    :param legend: label for each model
    :return: matplotlib figure instance
    """
    if not histories:
        return None

    fig = plt.figure(figsize=figsize, layout="tight")
    fig.suptitle(title)

    linestyles = iter(LINE_STYLES.values())
    for history in histories:
        linestyle = next(linestyles)
        epochs = list(history.keys())
        if len(epochs) > 1:
            avg_costs = []
            for epoch in epochs:
                batches = history[epoch].keys()
                avg_batch_costs = []
                for batch in batches:
                    avg_batch_cost = np.mean(history[epoch][batch])
                    avg_batch_costs.append(avg_batch_cost)
                avg_costs.append(sum(avg_batch_costs) / len(batches))
            plt.plot(
                range(len(epochs)),
                np.array(avg_costs),
                alpha=alpha,
                color="k",
                linewidth=2,
                linestyle=linestyle,
            )
            plt.xlabel("epoch", fontsize=fontsize)
            plt.ylabel("avg cost", fontsize=fontsize)
        elif len(history["epoch_0"]) > 1:
            avg_cost = []
            batches = history["epoch_0"].keys()
            for batch in batches:
                avg_cost.append(np.mean(history["epoch_0"][batch]))
            plt.plot(
                range(len(batches)),
                avg_cost,
                alpha=alpha,
                color="k",
                linewidth=2,
                linestyle=linestyle,
            )
            plt.xlabel("batch", fontsize=fontsize)
            plt.ylabel("avg cost", fontsize=fontsize)
        else:
            cost = history["epoch_0"]["batch_0"]
            plt.plot(
                range(len(cost)),
                cost,
                alpha=alpha,
                color="k",
                linewidth=2,
                linestyle=linestyle,
            )
            plt.xlabel("iteration", fontsize=fontsize)
            plt.ylabel("cost", fontsize=fontsize)
    ax = plt.gca()
    if legend:
        ax.legend(legend)
    ax.set_yscale("log")
    plt.close(fig)
    return fig


@requires_matplotlib
def residuals_by_predicted(
    y_pred: np.ndarray,
    y_true: np.ndarray,
    percent_residuals: bool = False,
    ax: Union[plt.Axes, None] = None,  # noqa: ANN401
    figsize: Tuple[float, float] = (3.25, 3),
    title: str = "",
    fontsize: int = 9,
    alpha: float = 1.0,
) -> plt.Figure:  # noqa: ANN401
    """Create residual by predicted plot for a single response.

    :param y_pred: predicted values, array of shape (m,)
    :param y_true: true values, array of shape (m,)
    :param percent_residuals: shows residuals as percentages if True
    :param ax: the matplotlib axes on which to plot the data
    :param figsize: figure size
    :param title: title of figure
    :param fontsize: text size
    :param alpha: transparency of dots (between 0 and 1)
    :return: matplotlib figure instance
    """
    if y_pred.ndim > 1:
        if y_pred.ndim == 2 and 1 in y_pred.shape:
            pass  # just needs to be unraveled
        else:
            raise ValueError(
                f"Expected one dimensional array, "
                f"but y_pred has {y_pred.ndim} dimensions."
            )
    if y_true.ndim > 1 and 1 not in y_true.shape:
        if y_true.ndim == 2 and 1 in y_true.shape:
            pass  # just needs to be unraveled
        else:
            raise ValueError(
                f"Expected one dimensional array, "
                f"but y_true has {y_true.ndim} dimensions."
            )
    fig = plt.figure(figsize=figsize, layout="tight")
    if not ax:
        spec = fig.add_gridspec(ncols=1, nrows=1)
        ax = fig.add_subplot(spec[0, 0])
    y_true = y_true.ravel()
    y_pred = y_pred.ravel()
    if percent_residuals:
        residuals = 100 * ((y_pred - y_true) / (y_true + 1e-12)).ravel()
    else:
        residuals = y_pred - y_true
    avg_error = residuals.mean()
    std_error = residuals.std()
    ax.axhline(y=avg_error, color="k", linestyle="-", linewidth=2)
    ax.axhline(y=avg_error + std_error, color="k", linestyle=":", linewidth=2)
    ax.axhline(y=avg_error - std_error, color="k", linestyle=":", linewidth=2)
    ax.scatter(y_pred, residuals, color="k", alpha=alpha)
    ax.axhline(y=0, color="r", linestyle=":")
    ax.set_title(title, fontsize=fontsize)
    if percent_residuals:
        ax.set_ylabel("Residuals (%)", fontsize=fontsize)
    else:
        ax.set_ylabel("Residuals", fontsize=fontsize)
    ax.set_xlabel("Predicted", fontsize=fontsize)
    ax.grid(True)
    ax.legend([f"avg = {avg_error:.3f}", f"std = {std_error:.3f}"], fontsize=fontsize)
    plt.close(fig)
    return fig


@requires_matplotlib
def goodness_of_fit(
    y_true: np.ndarray,
    y_pred: np.ndarray,
    percent_residuals: bool = False,
    figsize: Tuple[float, float] = (6.5, 3),
    fontsize: int = 9,
    alpha: float = 1.0,
    title: str = "",
) -> plt.Figure:  # noqa: ANN401
    """Create 'residual by predicted' and 'actual by predicted' plots.

    :param y_true: true values, array of shape (m,)
    :param y_pred: predicted values, array of shape (m,)
    :param percent_residuals: shows residuals as percentages if True
    :param figsize: figure size
    :param fontsize: text size
    :param alpha: transparency of dots (between 0 and 1)
    :param title: title of figure
    :return: matplotlib figure instance
    """
    if title is None:
        title = ""
    r2 = r_square(y_pred, y_true).squeeze()
    fig = plt.figure(figsize=figsize, layout="tight")
    fig.suptitle(title + f" (R-Squared = {r2:.3f})")
    spec = fig.add_gridspec(ncols=2, nrows=1)
    ax0 = fig.add_subplot(spec[0, 0])
    actual_by_predicted(
        ax=ax0,
        y_pred=y_pred,
        y_true=y_true,
        fontsize=fontsize,
        alpha=alpha,
    )
    ax1 = fig.add_subplot(spec[0, 1])
    residuals_by_predicted(
        ax=ax1,
        y_pred=y_pred,
        y_true=y_true,
        percent_residuals=percent_residuals,
        fontsize=fontsize,
        alpha=alpha,
    )
    plt.close(fig)
    return fig


@requires_matplotlib
def sensitivity_profile(
    ax: plt.Axes,  # noqa: ANN401
    x0: np.ndarray,
    y0: np.ndarray,
    x_pred: np.ndarray,
    y_pred: Union[np.ndarray, List[np.ndarray]],
    x_true: Union[np.ndarray, None] = None,
    y_true: Union[np.ndarray, None] = None,
    alpha: float = 1.0,
    xlabel: str = "x",
    ylabel: str = "y",
    legend: Union[List[str], None] = None,
    figsize: Tuple[float, float] = (6.5, 3),
    fontsize: int = 9,
    show_cursor: bool = True,
) -> plt.Figure:  # noqa: ANN401
    """Plot sensitivity profile for a single input, single output.

    :param ax: the matplotlib axes on which to plot the data
    :param x0: point at which the profile is centered, array of shape
        (1,)
    :param y0: model evaluated as x0, list of arrays of shape (1,)
    :param x_pred: input values for prediction, an array of shape (m,)
    :param y_pred: predicted output values for each model, list of
        arrays of shape (m,)
    :param x_true: inputs value of actual data, array of shape (m, n_x)
    :param y_true: output values of actual data. An array of shape (m,)
    :param alpha: transparency of dots (between 0 and 1)
    :param xlabel: label of x-axis
    :param ylabel: label of y-axis
    :param legend: legend name of each model
    :param figsize: figure size
    :param fontsize: text size
    :param show_cursor: show x0 as a red dot (or not)
    :return: matplotlib figure instance
    """
    fig = plt.figure(figsize=figsize, layout="tight")
    if not ax:
        spec = fig.add_gridspec(ncols=1, nrows=1)
        ax = fig.add_subplot(spec[0, 0])
    if not isinstance(y_pred, list):
        y_pred = [y_pred]
    if legend is None:
        legend = []
    x0 = x0.ravel()
    y0 = y0.ravel()
    x_pred = x_pred.ravel()
    linestyles = iter(LINE_STYLES.values())
    for array in y_pred:
        linestyle = next(linestyles)
        ax.plot(x_pred, array.ravel(), color="k", linestyle=linestyle, linewidth=2)
    if x_true is not None and y_true is not None:
        x_true = x_true.ravel()
        y_true = y_true.ravel()
        ax.scatter(x_true, y_true, color="k", alpha=alpha)
        legend.append("data")
    ax.legend(legend, fontsize=fontsize)
    if show_cursor:
        for n in range(y0.size):
            ax.scatter(x0, y0[n], color="r")
    ax.set_xlabel(xlabel, fontsize=fontsize)
    ax.set_ylabel(ylabel, fontsize=fontsize)
    ax.grid(True)
    plt.close(fig)
    return fig


@requires_matplotlib
def sensitivity_profiles(
    f: Union[Callable, List[Callable]],
    x_min: np.ndarray,
    x_max: np.ndarray,
    x0: Union[np.ndarray, None] = None,
    x_true: Union[np.ndarray, None] = None,
    y_true: Union[np.ndarray, None] = None,
    figsize: Tuple[float, float] = (3.25, 3),
    fontsize: int = 9,
    alpha: float = 1.0,
    title: str = "",
    xlabels: Union[List[str], None] = None,
    ylabels: Union[List[str], None] = None,
    legend: Union[List[str], None] = None,
    resolution: int = 100,
    show_cursor: bool = True,
) -> plt.Figure:  # noqa: ANN401
    """Plot grid of all outputs vs. all inputs evaluated at x0.

    :param f: callable function(s) for evaluating y_pred = f_pred(x)
    :param x0: point at which to evaluate profiles, array of shape (n_x, 1)
    :param x_true: inputs at which y_true is evaluated, array of shape (n_x, m)
    :param y_true: true values, array of shape (n_y, m)
    :param figsize: figure size
    :param fontsize: text size
    :param alpha: transparency of dots (between 0 and 1)
    :param title: title of figure
    :param xlabels: x-axis labels
    :param ylabels: y-axis labels
    resolution: line resolution
    :param legend: legend labels for each model
    :param show_cursor: show x0 as a red dot (or not)
    """
    funcs = f
    if not isinstance(funcs, list):
        funcs = [funcs]
    x_min = x_min.ravel()
    x_max = x_max.ravel()
    if x0 is None:
        x0 = 0.5 * (x_min + x_max).reshape((-1, 1))
    y0 = np.concatenate([func(x0) for func in funcs], axis=1)
    n_x = x0.shape[0]
    n_y = y0.shape[0]
    x_indices = range(n_x)
    y_indices = range(n_y)
    xlabels = xlabels if xlabels else [f"x_{i}" for i in x_indices]
    ylabels = ylabels if ylabels else [f"y_{i}" for i in y_indices]
    width, height = figsize
    fig = plt.figure(figsize=(n_x * width, height), layout="tight")
    fig.suptitle(title)
    spec = fig.add_gridspec(ncols=n_x, nrows=n_y)
    for i in x_indices:
        x_pred = np.tile(x0, (1, resolution))
        x_pred[i] = np.linspace(x_min[i], x_max[i], resolution)
        y_preds = []
        for func in funcs:
            y_pred = func(x_pred)
            y_preds.append(y_pred)
        for j in y_indices:
            sensitivity_profile(
                ax=fig.add_subplot(spec[j, i]),
                x0=x0[i],
                y0=y0[j],
                x_pred=x_pred[i],
                y_pred=[y_pred[j] for y_pred in y_preds],
                x_true=x_true[i] if x_true is not None else None,
                y_true=y_true[i] if y_true is not None else None,
                fontsize=fontsize,
                alpha=alpha,
                xlabel=xlabels[i],
                ylabel=ylabels[j],
                legend=legend,
                show_cursor=show_cursor,
            )
        plt.close(fig)
    return fig



"""Load trained JMP model."""

from pathlib import Path
from typing import List, Tuple, Union

import numpy as np

from ..model import NeuralNet


def _is_float(string: str) -> bool:
    """Check is string can be converted to float."""
    tokens = string.split(".")
    for token in tokens:
        if not token.replace("-", "").isnumeric():
            return False
    return True


def _count_repetitions(char: str, string: str) -> int:
    """Count number of times a character is repeated."""
    count = 0
    for c in string:
        if c == char:
            count += 1
    return count


def _is_empty(items: List[list]) -> bool:
    """Return True is all items are empty lists."""
    if not isinstance(items, list):
        return False
    if all([item == [] for item in items]):
        return True
    if all([_is_empty(item) for item in items]):
        return True
    return False


def _expand(tokens: List[str]) -> List[str]:
    """Expand w * (a * x + b) into (w * a * x + w * b)."""
    updated = []
    number_open_brackets = 0
    factor = 1.0
    skip = 0
    N = len(tokens)
    for i, token in enumerate(tokens):
        if skip > 0:
            skip -= 1
            continue
        number_open_brackets += _count_repetitions("(", token) - _count_repetitions(
            ")", token
        )
        if number_open_brackets == 0:
            if _is_float(token):
                if i < N - 2:
                    if tokens[i + 1] == "*":
                        if tokens[i + 2] == "(":
                            factor = float(token)
                            number_open_brackets += 1
                            skip = 2
                            continue
            if token == ")":
                continue
        if number_open_brackets == 1:
            if tokens[i + 1] == ")":
                factor = 1.0
        if _is_float(token) and number_open_brackets == 1:
            updated.append(str(float(token) * factor))
        else:
            updated.append(token)
    return updated


def _remove_unnecessary_brackets(tokens: List[str]) -> List[str]:
    """Remove situations such as ["(", "-1.33806951259538", ")"] where brackets
    aren't needed."""
    updated = []
    skip = 0
    N = len(tokens)
    for i, token in enumerate(tokens):
        if skip > 0:
            skip -= 1
            continue
        if i < N - 2:
            if token == "(" and _is_float(tokens[i + 1]) and tokens[i + 2] == ")":
                updated.append(tokens[i + 1])
                skip = 2
                continue
        updated.append(token)
    return updated


def _get_node_parameters(
    tokens: List[str],
) -> Tuple[List[float], List[float], List[List[str]]]:
    """Return bias, weights, and activations associated with node in layer."""
    number_open_brackets = 0
    biases = []
    weights = []
    activations = []
    activation = []
    for i, token in enumerate(tokens):
        number_open_brackets += _count_repetitions("(", token) - _count_repetitions(
            ")", token
        )
        if number_open_brackets > 0:
            activation.append(token)
        if number_open_brackets == 0:
            if activation:
                expanded = _expand(
                    activation[1:]
                )  # remove "(" b/c only want inputs of Tanh(...)
                activations.append(expanded)
            activation = []
            if _is_float(token):
                if tokens[i + 1] == "+":
                    biases.append(float(token))  # there is only one per model in JMP
                elif tokens[i + 1] == "*":
                    weights.append(float(token))
    bias = [sum(biases)]
    return bias, weights, activations


def _get_layer_parameters(
    nodes: List[List[str]],
) -> Tuple[List[List[float]], List[List[float]], List[List[str]]]:
    """Return bias, weight, activations associated with current layer."""
    layer_biases = []
    layer_weights = []
    for node in nodes:  # for each node in layer
        (
            node_bias,
            node_weights,
            layer_activation_inputs,
        ) = _get_node_parameters(node)
        layer_biases.append(node_bias)
        layer_weights.append(node_weights)
    return layer_biases, layer_weights, layer_activation_inputs


def _get_tokens(equation: str) -> List[str]:
    """Parse equation into list of str."""
    equation = equation.replace("(", " ( ")
    equation = equation.replace(")", " ) ")
    equation = equation.replace("+", " + ")
    equation = equation.replace("*", " * ")
    return _remove_unnecessary_brackets(equation.split())


def _load_from_file(filename: Union[str, Path]) -> str:
    """Load JMP equation from local file."""
    with open(filename) as file:
        equation = "".join(file.readlines())
    return equation


def from_jmp(equation: Union[str, Path]) -> NeuralNet:
    """Load trained JMP model given formula.

    .. Note::
        Expected equation assumed to be obtained from JMP
        using the "Save Profile Formulas" method and copy/paste.

    .. Note::
        JMP yields a separate equation for each output. It does
        not provided a single equation that predicts all outputs
        at once. This function therefore yields NeuralNet objects
        that predict only a single output (consistent with JMP).

    .. Warning::
        Order of inputs matches order used in JMP. Burden is
        on user to keep track of variable ordering.

    :param equation: either the equation itself or a filename containing it
    :return: jenn.model.NeuralNet object preloaded with the JMP parameters
    """
    if isinstance(equation, Path) or "+" not in equation or "TanH" not in equation:
        equation = _load_from_file(equation)

    tokens = _get_tokens(equation)

    # Last layer
    (
        layer_bias,
        layer_weights,
        layer_activation_inputs,  # TanH(...) for each node in layer
    ) = _get_layer_parameters([tokens])

    # Initialize list of neural net parameters for each layer
    biases = [layer_bias]
    weights = [layer_weights]

    # Loop through previous layers
    while not _is_empty(layer_activation_inputs):
        (
            layer_bias,
            layer_weights,
            layer_activation_inputs,
        ) = _get_layer_parameters(nodes=layer_activation_inputs)
        biases.append(layer_bias)
        weights.append(layer_weights)

    # Parameters as arrays
    N = len(biases)  # number of layers
    b = [np.array(b) for b in reversed(biases)]
    W = [np.array(W) for W in reversed(weights)]

    # Load JMP parameters into NeuralNet object
    jmp_model = NeuralNet(
        layer_sizes=[W[0].shape[1]] + [W[i].shape[0] for i in range(N)],
        hidden_activation="tanh",
        output_activation="linear",
    )
    jmp_model.parameters.initialize()
    for i in range(1, N + 1):
        jmp_model.parameters.W[i][:] = W[i - 1]
        jmp_model.parameters.b[i][:] = b[i - 1]

    return jmp_model



"""Jenn module entry point."""

# Copyright (c) 2018 Steven H. Berguin
# Distributed under the terms of the MIT License.

from . import core, model, synthetic, utils

__version__ = "1.0.7"

__all__ = [
    "__version__",
    "core",
    "model",
    "utils",
    "synthetic",
]



"""Test training and prediction (including partials)."""
import jenn
import numpy as np

from ._utils import finite_difference


class TestSinusoid: 
    """Check model against 1D sinusoidal test function."""    
    
    def test_neural_net(self, m_train: int = 100, m_test: int = 30):
       """Train a neural net against 1D sinuidal data."""
       #########
       # Train #
       #########

       x_train, y_train, dydx_train = jenn.synthetic.Sinusoid.sample(0, m_train)
       nn = jenn.model.NeuralNet([1, 12, 1], 'tanh')
       nn.fit(x_train, y_train, 
              lambd=0.1, alpha=.01, max_iter=2000, is_normalize=True)

       #############################
       # Goodness of Fit: Training #
       #############################

       expected = y_train
       computed = nn.predict(x_train)
       score = jenn.utils.metrics.r_square(expected, computed)
       assert np.all(score > 0.95), f'r-square = {score} < 0.95'

       expected = dydx_train
       computed = nn.predict_partials(x_train)
       score = jenn.utils.metrics.r_square(expected, computed)
       assert np.all(score > 0.95), f'r-square = {score} < 0.95'

       expected = finite_difference(nn.predict, x_train)
       computed = nn.predict_partials(x_train)
       score = jenn.utils.metrics.r_square(expected, computed)
       assert np.all(score > 0.95), f'r-square = {score} < 0.95'

       ############################
       # Goodness of Fit: Testing #
       ############################

       x_test, y_test, dydx_test = jenn.synthetic.Sinusoid.sample(m_test)

       expected = y_test
       computed = nn.predict(x_test)
       assert np.all(jenn.utils.metrics.r_square(expected, computed) > 0.95), f'r-square = {score} < 0.95'

       expected = dydx_test
       computed = nn.predict_partials(x_test)
       assert np.all(jenn.utils.metrics.r_square(expected, computed) > 0.90), f'r-square = {score} < 0.90'

       expected = finite_difference(nn.predict, x_test)
       computed = nn.predict_partials(x_test)
       assert np.allclose(expected, computed, atol=1e-4) 
       
    def test_gradient_enhanced_neural_net(self, m_train: int = 4, m_test: int = 30):
       """Verify that gradient-enhancement succeeds in training a neural net
       against 1D sinuidal data when insufficient number of points is provided
       for a regular neural net to succeed."""
       x_train, y_train, dydx_train = jenn.synthetic.Sinusoid.sample(0, m_train)
       x_test, y_test, dydx_test = jenn.synthetic.Sinusoid.sample(m_test)

       #########
       # Train # (regular neural net)
       #########

       nn = jenn.model.NeuralNet([1, 12, 1], 'tanh')
       nn.fit(x_train, y_train,
              lambd=0.1, alpha=.01, max_iter=2000, is_normalize=True)

       #############################
       # Goodness of Fit: Training # (regular neural net should generalize badly)
       #############################

       expected = y_test
       computed = nn.predict(x_test)
       score = jenn.utils.metrics.r_square(expected, computed)
       assert np.all(score < 0.5), f'r-square = {score} > 0.5'  

       #########
       # Train # (gradient-enhanced neural net)
       #########

       genn = jenn.model.NeuralNet([1, 12, 1], 'tanh')
       genn.fit(x_train, y_train, dydx_train,
              lambd=0.1, gamma=1.0, alpha=0.05, max_iter=500, is_normalize=True)

       #############################
       # Goodness of Fit: Training #
       #############################

       expected = y_train
       computed = genn.predict(x_train)
       score = jenn.utils.metrics.r_square(expected, computed)
       assert np.all(score > 0.95), f'r-square = {score} < 0.95'

       expected = dydx_train
       computed = genn.predict_partials(x_train)
       score = jenn.utils.metrics.r_square(expected, computed)
       assert np.all(score > 0.95), f'r-square = {score} < 0.95'

       ############################
       # Goodness of Fit: Testing #
       ############################

       expected = y_test
       computed = genn.predict(x_test)
       assert np.all(jenn.utils.metrics.r_square(expected, computed) > 0.9), f'r-square = {score} < 0.9'

       expected = dydx_test
       computed = genn.predict_partials(x_test)
       assert np.all(jenn.utils.metrics.r_square(expected, computed) > 0.9), f'r-square = {score} < 0.9'


class TestRastrigin: 
    """Check model against 2D sinusoidal test function."""    

    @classmethod
    def test_neural_net(self): 
        """TODO"""
        pass 

    @classmethod
    def test_gradient_enhanced_neural_net(self): 
        """TODO"""
        pass 


"""Propagation.
==============

This module contains the critical functionality to propagate information forward and backward through the neural net."""

from typing import Tuple, Union

import numpy as np

from .activation import ACTIVATIONS
from .cache import Cache
from .data import Dataset
from .parameters import Parameters


def eye(n: int, m: int) -> np.ndarray:
    """Copy identify matrix of shape (n, n) m times."""
    eye = np.eye(n, dtype=float)
    return np.repeat(eye.reshape((n, n, 1)), m, axis=2)


def first_layer_forward(X: np.ndarray, cache: Union[Cache, None] = None) -> None:
    """Compute input layer activations (in place).

    :param X: training data inputs, array of shape (n_x, m)
    :param cache: neural net cache that stores neural net quantities
        computed during forward prop for each layer, so they can be
        accessed during backprop to avoid re-computing them
    """
    X = X.astype(float, copy=False)
    if cache is not None:
        cache.A[0][:] = X


def first_layer_partials(X: np.ndarray, cache: Union[Cache, None]) -> None:
    """Compute input layer partial (in place).

    :param X: training data inputs, array of shape (n_x, m)
    :param cache: neural net cache that stores neural net quantities
        computed during forward prop for each layer, so they can be
        accessed during backprop to avoid re-computing them
    """
    X = X.astype(float, copy=False)
    if cache is not None:
        n_x, m = X.shape
        cache.A_prime[0][:] = eye(n_x, m)


def next_layer_partials(layer: int, parameters: Parameters, cache: Cache) -> np.ndarray:
    """Compute j^th partial in place for one layer (in place).

    :param layer: index of current layer.
    :param parameters: object that stores neural net parameters for each
        layer
    :param cache: neural net cache that stores neural net quantities
        computed during forward prop for each layer, so they can be
        accessed during backprop to avoid re-computing them
    """
    s = layer
    r = layer - 1
    W = parameters.W[layer]
    g = ACTIVATIONS[parameters.a[layer]]
    cache.G_prime[s][:] = g.first_derivative(cache.Z[s], cache.A[s])
    for j in range(parameters.n_x):
        cache.Z_prime[s][:, j, :] = np.dot(W, cache.A_prime[r][:, j, :])
        cache.A_prime[s][:, j, :] = cache.G_prime[s] * np.dot(
            W, cache.A_prime[r][:, j, :]
        )
    return cache.A_prime[s]


def next_layer_forward(layer: int, parameters: Parameters, cache: Cache) -> None:
    """Propagate forward through one layer (in place).

    :param layer: index of current layer.
    :param parameters: object that stores neural net parameters for each
        layer
    :param cache: neural net cache that stores neural net quantities
        computed during forward prop for each layer, so they can be
        accessed during backprop to avoid re-computing them
    """
    s = layer
    r = layer - 1
    W = parameters.W[s]
    b = parameters.b[s]
    g = ACTIVATIONS[parameters.a[s]]
    Z = cache.Z[s]
    A = cache.A[s]
    np.dot(W, cache.A[r], out=Z)
    Z += b
    g.evaluate(Z, A)


def model_partials_forward(
    X: np.ndarray, parameters: Parameters, cache: Cache
) -> Tuple[np.ndarray, np.ndarray]:
    """Propagate forward in order to predict reponse(r) and partial(r).

    :param X: training data inputs, array of shape (n_x, m)
    :param parameters: object that stores neural net parameters for each
        layer
    :param cache: neural net cache that stores neural net quantities
        computed during forward prop for each layer, so they can be
        accessed during backprop to avoid re-computing them
    """
    first_layer_forward(X, cache)
    first_layer_partials(X, cache)
    for layer in parameters.layers[1:]:  # type: ignore[index]
        next_layer_forward(layer, parameters, cache)
        next_layer_partials(layer, parameters, cache)
    return cache.A[-1], cache.A_prime[-1]


def model_forward(X: np.ndarray, parameters: Parameters, cache: Cache) -> np.ndarray:
    """Propagate forward in order to predict reponse(r).

    :param X: training data inputs, array of shape (n_x, m)
    :param parameters: object that stores neural net parameters for each
        layer
    :param cache: neural net cache that stores neural net quantities
        computed during forward prop for each layer, so they can be
        accessed during backprop to avoid re-computing them
    """
    first_layer_forward(X, cache)
    for layer in parameters.layers[1:]:  # type: ignore[index]
        next_layer_forward(layer, parameters, cache)
    return cache.A[-1]


def partials_forward(X: np.ndarray, parameters: Parameters, cache: Cache) -> np.ndarray:
    """Propagate forward in order to predict partial(r).

    :param X: training data inputs, array of shape (n_x, m)
    :param parameters: object that stores neural net parameters for each
        layer
    :param cache: neural net cache that stores neural net quantities
        computed during forward prop for each layer, so they can be
        accessed during backprop to avoid re-computing them
    """
    return model_partials_forward(X, parameters, cache)[-1]


def last_layer_backward(cache: Cache, data: Dataset) -> None:
    """Propagate backward through last layer (in place).

    :param cache: neural net cache that stores neural net quantities
        computed during forward prop for each layer, so they can be
        accessed during backprop to avoid re-computing them
    :param data: object containing training and associated metadata
    """
    cache.dA[-1][:] = data.Y_weights * (cache.A[-1] - data.Y)
    if data.J is not None:
        cache.dA_prime[-1][:] = data.J_weights * (cache.A_prime[-1] - data.J)


def next_layer_backward(
    layer: int, parameters: Parameters, cache: Cache, data: Dataset, lambd: float
) -> None:
    """Propagate backward through next layer (in place).

    :param layer: index of current layer.
    :param parameters: object that stores neural net parameters for each
        layer
    :param cache: neural net cache that stores neural net quantities
        computed during forward prop for each layer, so they can be
        accessed during backprop to avoid re-computing them
    :param data: object containing training and associated metadata
    :param lambd: coefficient that multiplies regularization term in
        cost function
    """
    s = layer
    r = layer - 1
    g = ACTIVATIONS[parameters.a[s]]
    g.first_derivative(cache.Z[s], cache.A[s], cache.G_prime[s])
    np.dot(cache.G_prime[s] * cache.dA[s], cache.A[r].T, out=parameters.dW[s])
    parameters.dW[s] /= data.m
    parameters.dW[s] += lambd / data.m * parameters.W[s]
    np.sum(cache.G_prime[s] * cache.dA[s], axis=1, keepdims=True, out=parameters.db[s])
    parameters.db[s] /= data.m
    np.dot(parameters.W[s].T, cache.G_prime[s] * cache.dA[s], out=cache.dA[r])


def gradient_enhancement(
    layer: int,
    parameters: Parameters,
    cache: Cache,
    data: Dataset,
) -> None:
    """Add gradient enhancement to backprop (in place).

    :param layer: index of current layer.
    :param parameters: object that stores neural net parameters for each
        layer
    :param cache: neural net cache that stores neural net quantities
        computed during forward prop for each layer, so they can be
        accessed during backprop to avoid re-computing them
    :param data: object containing training and associated metadata
    """
    if data.J is None:
        return
    if np.all(data.J_weights == 0.0):
        return
    s = layer
    r = layer - 1
    g = ACTIVATIONS[parameters.a[s]]
    cache.G_prime_prime[s][:] = g.second_derivative(
        cache.Z[s], cache.A[s], cache.G_prime[s]
    )
    coefficient = 1 / data.m
    for j in range(parameters.n_x):
        parameters.dW[s] += coefficient * (
            np.dot(
                cache.dA_prime[s][:, j, :]
                * cache.G_prime_prime[s]
                * cache.Z_prime[s][:, j, :],
                cache.A[r].T,
            )
            + np.dot(
                cache.dA_prime[s][:, j, :] * cache.G_prime[s],
                cache.A_prime[r][:, j, :].T,
            )
        )
        parameters.db[s] += coefficient * np.sum(
            cache.dA_prime[s][:, j, :]
            * cache.G_prime_prime[s]
            * cache.Z_prime[s][:, j, :],
            axis=1,
            keepdims=True,
        )
        cache.dA[r] += np.dot(
            parameters.W[s].T,
            cache.dA_prime[s][:, j, :]
            * cache.G_prime_prime[s]
            * cache.Z_prime[s][:, j, :],
        )
        cache.dA_prime[r][:, j, :] = np.dot(
            parameters.W[s].T, cache.dA_prime[s][:, j, :] * cache.G_prime[s]
        )


def model_backward(
    data: Dataset,
    parameters: Parameters,
    cache: Cache,
    lambd: float = 0.0,
) -> None:
    """Propagate backward through all layers (in place).

    :param parameters: object that stores neural net parameters for each
        layer
    :param cache: neural net cache that stores neural net quantities
        computed during forward prop for each layer, so they can be
        accessed during backprop to avoid re-computing them
    :param data: object containing training and associated metadata
    :param lambd: regularization coefficient to avoid overfitting
        [defaulted to zero] (optional)
    """
    last_layer_backward(cache, data)
    for layer in reversed(parameters.layers):  # type: ignore[call-overload]
        if layer > 0:
            next_layer_backward(layer, parameters, cache, data, lambd)
            gradient_enhancement(layer, parameters, cache, data)



"""Test that cost function is working."""
import jenn
import numpy as np


def test_least_squares():
    """Test that least squares cost function evaluates to known answers."""
    x, y, dydx = jenn.synthetic.Sinusoid.sample(100)

    parameters = jenn.core.parameters.Parameters(layer_sizes=[2, 2, 1])
    parameters.initialize()
    
    data = jenn.core.data.Dataset(x, y, dydx)
    cost = jenn.core.cost.Cost(data, parameters, lambd=0.0)

    # Verify that cost is zero when prediction is perfect
    assert cost.evaluate(y, dydx) == 0

    # Verify that cost is non-zero when prediction is imperfect
    eps = 2 * np.random.rand()
    assert cost.evaluate(y + eps, dydx + eps) != 0

    # TODO: check prediction returns known result


def test_regularization():
    pass  # TODO 




"""Check that optimizer by testing it against 
standard cannonical test functions."""
import pytest 
import numpy as np
import jenn
from importlib.util import find_spec

if find_spec("matplotlib"):
    from matplotlib import pyplot as plt
    MATPLOTLIB_INSTALLED = True
else:
    MATPLOTLIB_INSTALLED = False

class TestLineSearch: 
    """Check that line search is working 
    using 1D parabola as test function."""

    @pytest.fixture
    def line_search(self) -> jenn.core.optimization.Backtracking:
        """Return backtracking line search using gradient descent."""
        return jenn.core.optimization.Backtracking(
            update=jenn.core.optimization.GD()
        )

    def test_finds_minimum(self, line_search: jenn.core.optimization.Backtracking):
        """Test that line search finds the minimum when learning rate step size
        is large enough to include minimum in search radius."""
        center = 1.0  # center parabola at this number 
        f = lambda x: jenn.synthetic.Parabola.evaluate(x, x0=center)
        dfdx = lambda x: jenn.synthetic.Parabola.first_derivative(x, x0=center)
        x0 = np.array([center - 1]).reshape((1, 1))  # approach from left 
        assert line_search(x0, dfdx(x0), f, learning_rate=2.0) == center
        x0 = np.array([center + 1]).reshape((1, 1))  # approach from right 
        assert line_search(x0, dfdx(x0), f, learning_rate=2.0) == center

    def test_finds_minbound(self, line_search: jenn.core.optimization.Backtracking):
        """Test that line search finds the minimum bound when learning rate step size
        is not large enough to include minimum in search radius."""
        f = lambda x: jenn.synthetic.Parabola.evaluate(x, x0=0.0)
        dfdx = lambda x: jenn.synthetic.Parabola.first_derivative(x, x0=0.0)
        x0 = np.array([1]).reshape((1, 1))  # approach from right 
        assert line_search(x0, dfdx(x0), f, learning_rate=0.1) == 0.8
        x0 = np.array([-1]).reshape((1, 1))  # approach from left 
        assert line_search(x0, dfdx(x0), f, learning_rate=0.1) == -0.8


class TestUpdate: 
    """Test parameter update using a simple 
    linear function."""

    def test_GD(self):
        """Test gradient descent using simple linear function."""
        x0 = np.array([5, 10]).reshape((1, -1))
        dydx = jenn.synthetic.Linear.first_derivative(x0)
        update = jenn.core.optimization.GD()
        x = update(x0, dydx, alpha=1).squeeze()
        assert np.allclose(x, np.array([4, 9]))

    def test_ADAM(self):
        """Test ADAM using simple linear function."""
        x0 = np.array([5, 10]).reshape((1, -1))
        dydx = jenn.synthetic.Linear.first_derivative(x0)
        update = jenn.core.optimization.ADAM()
        x = update(x0, dydx, alpha=1).squeeze()
        assert np.allclose(x, np.array([4, 9]))


class TestOptimizer: 
    """Test optimizer using banana Rosenbrock function."""
    
    @classmethod
    def test_rosenbrock(
            cls, 
            alpha: float = 0.05, 
            max_iter: int = 1000,
            is_adam: bool = True, 
            is_plot: bool = False,
        ):
        """Check that optimizer yields correct answer for rosenbrock function."""

        # Initial guess
        x0 = np.array([1.25, -1.75]).reshape((2, 1))

        # Test function
        f = jenn.synthetic.Rosenbrock.evaluate
        dfdx = jenn.synthetic.Rosenbrock.first_derivative

        # Optimization
        if is_adam:
            opt = jenn.core.optimization.ADAMOptimizer()
        else:
            opt = jenn.core.optimization.GDOptimizer()

        xf = opt.minimize(x0, f, dfdx, alpha=alpha, max_iter=max_iter)

        # For plotting contours
        lb = -2.
        ub = 2.
        m = 100
        x1 = np.linspace(lb, ub, m)
        x2 = np.linspace(lb, ub, m)
        X1, X2 = np.meshgrid(x1, x2)
        Y = np.zeros(X1.shape)
        for i in range(0, m):
            for j in range(0, m):
                X = np.array([
                    [X1[i, j]],
                    [X2[i, j]],
                ])
                Y[i, j] = f(X)

        if not MATPLOTLIB_INSTALLED:
            # raise ImportError("Matplotlib must be installed.")
            return 

        if is_plot:
            x1_his = np.array([x[0] for x in opt.vars_history]).squeeze()
            x2_his = np.array([x[1] for x in opt.vars_history]).squeeze()
            plt.plot(x1_his, x2_his)
            plt.plot(x0[0], x0[1], '+', ms=15)
            plt.plot(xf[0], xf[1], 'o')
            plt.plot(np.array([1.]), np.array([1.]), 'x')
            plt.legend(
                ['history', 'initial guess', 'predicted optimum', 'true optimum'])
            plt.contour(X1, X2, Y, 50, cmap='RdGy')
            if is_adam:
                plt.title('ADAM')
            else:
                plt.title('GD')
            plt.show()

        # Close to optimum, the slope is nearly zero and the optimizer really
        # struggles to get to the exact optimum. +/- 0.2 is actually very close
        # to optimality. Turn on the contour plots to see.
        assert np.allclose(xf[0], 1.0, atol=0.2)
        assert np.allclose(xf[1], 1.0, atol=0.2)


if __name__ == "__main__":
    TestOptimizer.test_rosenbrock(alpha=0.1, max_iter=1_000, is_adam=True, is_plot=True)



"""Cost Function.
=================

This module contains class and methods to efficiently 
compute the neural net cost function used for training. 
It is a modified version of the Least Squared Estimator (LSE), 
augmented with a penalty function for regularization and another 
term which accounts for Jacobian prediction error. See
`paper`_ for details and notation. 
"""  # noqa W291

from typing import List, Union

import numpy as np

from .data import Dataset
from .parameters import Parameters


class SquaredLoss:
    r"""Least Squares Estimator.

    :param Y_true: training data outputs :math:`Y \in \mathbb{R}^{n_y
        \times m}`
    :param Y_weights: weights by which to prioritize data points
        (optional)
    """

    def __init__(
        self, Y_true: np.ndarray, Y_weights: Union[np.ndarray, float] = 1.0
    ) -> None:
        self.Y_true = Y_true
        self.Y_error = np.zeros(Y_true.shape)  # preallocate to save resources
        self.Y_weights = np.ones(Y_true.shape) * Y_weights
        self.n_y, self.m = Y_true.shape

    def evaluate(self, Y_pred: np.ndarray) -> np.float64:
        r"""Compute least squares estimator of the states in place.

        :param Y_pred: predicted outputs :math:`A^{[L]} \in
            \mathbb{R}^{n_y \times m}`
        """
        self.Y_error[:, :] = Y_pred - self.Y_true
        self.Y_error *= np.sqrt(self.Y_weights)
        cost = 0
        for j in range(0, self.n_y):
            cost += np.dot(self.Y_error[j], self.Y_error[j].T)
        return np.float64(cost)


class GradientEnhancement:
    r"""Least Squares Estimator for partials.

    :param J_true: training data jacobian :math:`Y^{\prime} \in
        \mathbb{R}^{n_y \times m}`
    :param J_weights: weights by which to prioritize partials (optional)
    """

    def __init__(
        self, J_true: np.ndarray, J_weights: Union[np.ndarray, float] = 1.0
    ) -> None:
        self.J_true = J_true
        self.J_error = np.zeros(J_true.shape)
        self.J_weights = np.ones(J_true.shape) * J_weights
        self.n_y, self.n_x, self.m = J_true.shape

    def evaluate(self, J_pred: np.ndarray) -> np.float64:
        r"""Compute least squares estimator for the partials.

        :param J_pred: predicted Jacobian :math:`A^{\prime[L]} \in
            \mathbb{R}^{n_y \times n_x \times m}`
        """
        self.J_error[:, :, :] = self.J_weights * (J_pred - self.J_true)
        self.J_error *= np.sqrt(self.J_weights)
        cost = 0.0
        for k in range(0, self.n_y):
            for j in range(0, self.n_x):
                dot_product = np.dot(self.J_error[k, j], self.J_error[k, j].T)
                cost += np.squeeze(dot_product)
        return np.float64(cost)


class Regularization:
    """Compute regularization penalty."""

    def __init__(self, weights: List[np.ndarray], lambd: float = 0.0) -> None:
        r"""Compute L2 norm penalty.

        :param weights: neural parameters :math:`W^{[l]} \in
            \mathbb{R}^{n^{[l]} \times n^{[l-1]}}` associated with each
            layer
        """
        self.weights = weights
        self.lambd = lambd

    def evaluate(
        self,
    ) -> float:
        r"""Compute L2 norm penalty.

        :param weights: neural parameters :math:`W^{[l]} \in
            \mathbb{R}^{n^{[l]} \times n^{[l-1]}}` associated with each
            layer
        :param lambd: regularization coefficient :math:`\lambda \in
            \mathbb{R}` (hyperparameter to be tuned)
        """
        penalty = 0.0
        if self.lambd > 0.0:
            for W in self.weights:
                penalty += np.sum(np.square(W)).squeeze()
            return self.lambd * penalty
        return 0.0


class Cost:
    r"""Neural Network cost function.

    :param data: Dataset object containing training data (and associated
        metadata)
    :param parameters: object containing neural net parameters (and
        associated metadata) for each layer
    :param lambd: regularization coefficient to avoid overfitting
    """

    def __init__(
        self,
        data: Dataset,
        parameters: Parameters,
        lambd: float = 0.0,
    ) -> None:
        self.data = data
        self.parameters = parameters
        self.squared_loss = SquaredLoss(data.Y, data.Y_weights)
        self.regularization = Regularization(parameters.W, lambd)
        if data.J is not None:  # noqa: PLR2004
            self.gradient_enhancement = GradientEnhancement(data.J, data.J_weights)

    def evaluate(
        self, Y_pred: np.ndarray, J_pred: Union[np.ndarray, None] = None
    ) -> np.float64:
        r"""Evaluate cost function.

        :param Y_pred: predicted outputs :math:`A^{[L]} \in
            \mathbb{R}^{n_x \times m}`
        :param J_pred: predicted Jacobian :math:`A^{\prime[L]} \in
            \mathbb{R}^{n_y \times n_x \times m}`
        """
        cost = self.squared_loss.evaluate(Y_pred)
        if J_pred is not None and hasattr(self, "gradient_enhancement"):
            cost += self.gradient_enhancement.evaluate(J_pred)
        cost += self.regularization.evaluate()
        cost *= 0.5 / self.data.m
        return cost



"""Cache.
=========

This module defines a convenience class to  all quantities 
computed during forward propagation, so they don't have to be 
recomputed again during backward propgation. See
`paper`_ for details and notation.
"""  # noqa: W291

from typing import List

import numpy as np


class Cache:
    r"""Neural net cache.

    A cache s neural net quantities computed during
    forward prop for each layer, so they don't have to be
    recomputed again during backprop. This makes the algorithm faster.

    .. warning::
        The attributes of this class are not protected. It's possible
        to overwrite them instead of updating them in place. To ensure
        that an array is updated in place, use the numpy `[:]` syntax:

        .. code-block:: python

            cache = Cache(shapes)
            layer_1_activations = cache.A[1]
            layer_1_activations[:] = new_array_values  # note [:]

    .. note::
        The variables and their symbols refer to the theory in the companion
        `paper`_ for this library.

    :param layer_sizes: number of nodes in each layer (including input/output layers)
    :param m: number of examples (used to preallocate arrays)

    :ivar Z:  :math:`Z^{[l]} \in \mathbb{R}^{n^{[l]}\times m}~\forall~ l = 1 \dots L`
    :vartype Z: List[numpy.ndarray]

    :ivar Z_prime:  :math:`{Z^\prime}^{[l]} \in \mathbb{R}^{n^{[l]}\times n_x \times m}~\forall~ l = 1 \dots L`
    :vartype Z_prime: List[numpy.ndarray]

    :ivar A:  :math:`A^{[l]} = g(Z^{[l]}) \in \mathbb{R}^{n^{[l]} \times m}~\forall~ l = 1 \dots L`
    :vartype A: List[numpy.ndarray]

    :ivar A_prime:  :math:`{A^\prime}^{[l]} = g^\prime(Z^{[l]})Z^{\prime[l]} \in \mathbb{R}^{n^{[l]}\times n_x \times m}`
    :vartype A_prime: List[numpy.ndarray]

    :ivar G_prime:  :math:`G^{\prime} = g^{\prime}(Z^{[l]}) \in \mathbb{R}^{n^{[l]} \times m}~\forall~ l = 1 \dots L`
    :vartype G_prime: List[numpy.ndarray]

    :ivar G_prime_prime:  :math:`G^{\prime\prime} = g^{\prime\prime}(Z^{[l]}) \in \mathbb{R}^{n^{[l]} \times m}`
    :vartype G_prime_prime: List[numpy.ndarray]

    :ivar dA: :math:`{\partial \mathcal{J}}/{dA^{[l]}}  \in \mathbb{R}^{n^{[l]} \times m}~\forall~ l = 1 \dots L`
    :vartype dA: List[numpy.ndarray]

    :ivar dA_prime: :math:`{\partial \mathcal{J}}/{dA^{\prime[l]}}  \in \mathbb{R}^{n^{[l]} \times n_x \times m}~\forall~ l = 1 \dots L`
    :vartype dA: List[numpy.ndarray]
    """

    @property
    def m(self) -> int:
        """Return number of examples."""
        return int(self.A[0].shape[1])

    @property
    def n_x(self) -> int:
        """Return number of inputs."""
        return int(self.layer_sizes[0])

    @property
    def n_y(self) -> int:
        """Return number of outputs."""
        return int(self.layer_sizes[-1])

    def __init__(self, layer_sizes: List[int], m: int = 1):  # noqa: D107
        self.layer_sizes = layer_sizes
        self.Z: List[np.ndarray] = []  #  z = w a_prev + b
        self.Z_prime: List[np.ndarray] = []  #  z' = dz/dx[j] for all j = 1, .., n_x
        self.A: List[np.ndarray] = []  #  a = g(z)
        self.A_prime: List[np.ndarray] = []  #  a' = da/dx[j] for all j = 1, .., n_x
        self.G_prime: List[np.ndarray] = []  #  g' = da/dz
        self.G_prime_prime: List[np.ndarray] = []  #  g'' = d/dz( da/dz )
        self.dA: List[np.ndarray] = []
        self.dA_prime: List[np.ndarray] = []
        for n in self.layer_sizes:
            self.Z.append(np.zeros((n, m)))
            self.Z_prime.append(np.zeros((n, self.n_x, m)))
            self.G_prime.append(np.zeros((n, m)))
            self.G_prime_prime.append(np.zeros((n, m)))
            self.A.append(np.zeros((n, m)))
            self.A_prime.append(np.zeros((n, self.n_x, m)))
            self.dA.append(np.zeros((n, m)))
            self.dA_prime.append(np.zeros((n, self.n_x, m)))



"""Entry point for core modules."""

# Copyright (c) 2023 Steven H. Berguin
# Distributed under the terms of the MIT License.

from . import (
    activation,
    cache,
    cost,
    data,
    optimization,
    parameters,
    propagation,
    training,
)

__all__ = [
    "activation",
    "cache",
    "cost",
    "data",
    "optimization",
    "parameters",
    "propagation",
    "training",
]



"""Optimization.
================

.. ADAM: https://doi.org/10.48550/arXiv.1412.6980

This module implements gradient-based optimization using `ADAM`_. 
"""  # noqa W291

from abc import ABC, abstractmethod
from collections.abc import Callable
from typing import Union

import numpy as np


class Update(ABC):
    r"""Base class for line search.

    Update parameters :math:`\boldsymbol{x}` by taking a step along
    the search direction :math:`\boldsymbol{s}` according to
    :math:`\boldsymbol{x} := \boldsymbol{x} + \alpha
    \boldsymbol{s}`

    .. automethod:: __call__
    """  # noqa W291

    @abstractmethod
    def _update(
        self,
        params: np.ndarray,
        grads: np.ndarray,
        alpha: float,
    ) -> np.ndarray:
        raise NotImplementedError("To be implemented in subclass.")

    def __call__(
        self,
        params: np.ndarray,
        grads: np.ndarray,
        alpha: float,
    ) -> np.ndarray:
        r"""Take a single step along search direction.

        :param params: parameters :math:`x` to be updated
        :param grads: gradient :math:`\nabla_x f` of
            objective function :math:`f` w.r.t. each parameter
            :math:`x`
        :param alpha: learning rate :math:`\alpha`
        """
        return self._update(params, grads, alpha)


class GD(Update):
    r"""Take single step along the search direction using gradient descent.

    GD simply follows the steepest path according to
    :math:`\boldsymbol{x} := \boldsymbol{x} + \alpha \boldsymbol{s}`
    where :math:`\boldsymbol{s} = \nabla_x f`
    """

    def _update(
        self,
        params: np.ndarray,
        grads: np.ndarray,
        alpha: float,
    ) -> np.ndarray:
        return (params - alpha * grads).reshape(params.shape)


class ADAM(Update):
    r"""Take single step along the search direction as determined by `ADAM`_.

    Parameters :math:`\boldsymbol{x}` are updated according to
    :math:`\boldsymbol{x} := \boldsymbol{x} + \alpha \boldsymbol{s}`
    where :math:`\boldsymbol{s}` is determined by ADAM in such a way to
    improve efficiency. This is accomplished making use of previous
    information (see paper).

    :param beta_1: exponential decay rate of 1st moment vector
        :math:`\beta_1\in[0, 1)`
    :param beta_2: exponential decay rate of 2nd moment vector
        :math:`\beta_2\in[0, 1)`
    """

    def __init__(
        self,
        beta_1: float = 0.9,
        beta_2: float = 0.99,
    ):  # noqa D107
        self.beta_1 = beta_1
        self.beta_2 = beta_2

        self._v: Union[np.ndarray, None] = None
        self._s: Union[np.ndarray, None] = None
        self._t = 0
        self._grads: Union[np.ndarray, None] = None

    def _update(
        self,
        params: np.ndarray,
        grads: np.ndarray,
        alpha: float,
    ) -> np.ndarray:
        beta_1 = self.beta_1
        beta_2 = self.beta_2

        v = self._v
        s = self._s
        t = self._t

        if v is None:
            v = np.zeros(params.shape)
        if s is None:
            s = np.zeros(params.shape)

        a: list[int] = [id(x) for x in grads]
        b: list[int] = []
        if self._grads is not None:
            b = [id(x) for x in self._grads]

        if a != b:
            self._grads = grads
            t += 1  # only update for new search directions

        epsilon = np.finfo(float).eps  # small number to avoid division by zero
        v = beta_1 * v + (1.0 - beta_1) * grads
        s = beta_2 * s + (1.0 - beta_2) * np.square(grads)
        v_corrected = v / (1.0 - beta_1**t)
        s_corrected = s / (1.0 - beta_2**t) + epsilon

        x = params - alpha * v_corrected / np.sqrt(s_corrected)

        self._v = v
        self._s = s
        self._t = t

        return x.reshape(params.shape)


class LineSearch(ABC):
    r"""Take multiple steps of varying size by progressively varying
    :math:`\alpha` along the search direction.

    :param update: object that implements Update base class to update
        parameters according to :math:`\boldsymbol{x} :=
        \boldsymbol{x} + \alpha \boldsymbol{s}`

    .. automethod:: __call__
    """

    def __init__(
        self,
        update: Update,
    ):  # noqa D107
        self.update = update

    @abstractmethod
    def __call__(
        self,
        params: np.ndarray,
        grads: np.ndarray,
        cost: Callable,
        learning_rate: float,
    ) -> np.ndarray:
        r"""Take multiple steps along the search direction.

        :param params: parameters to be updated, array of shape (n,)
        :param grads: cost function gradient w.r.t. parameters, array of
            shape (n,)
        :param cost: cost function, array of shape (1,)
        :param learning_rate: initial step size :math:`\alpha`
        :return: new_params: updated parameters, array of shape (n,)
        """
        raise NotImplementedError


class Backtracking(LineSearch):
    r"""Search for optimum along a search direction.

    :param update: object that updates parameters according to :math:`\boldsymbol{x} := \boldsymbol{x} + \alpha \boldsymbol{s}`
    :param tau: amount by which to reduce :math:`\alpha := \tau \times \alpha` on each iteration
    :param tol: stop when cost function doesn't improve more than specified tolerance
    :param max_count: stop when line search iterations exceed maximum count specified

    .. automethod:: __call__
    """

    def __init__(
        self,
        update: Update,
        tau: float = 0.5,
        tol: float = 1e-6,
        max_count: int = 1_000,
    ):  # noqa D107
        super().__init__(update)
        self.tau = tau
        self.tol = tol
        self.max_count = max_count

    def __call__(
        self,
        params: np.ndarray,
        grads: np.ndarray,
        cost: Callable,
        learning_rate: float = 0.05,
    ) -> np.ndarray:
        r"""Take multiple "update" steps along search direction.

        :param params: parameters :math:`x` to be updated, array of
            shape (n,)
        :param grads: gradient :math:`\nabla_x f` of
            objective function :math:`f` w.r.t. each
            parameter, array of shape (n,)
        :param cost: objective function :math:`f`
        :param learning_rate: maximum allowed step size :math:`\alpha
            \le \alpha_{max}`
        :return: updated parameters :math:`x`, array of shape (n,)
        """
        tau = self.tau
        tol = self.tol
        x0 = self.update(params, grads, alpha=0)
        f0 = cost(x0)
        tau = max(0.0, min(1.0, tau))
        alpha = learning_rate
        x = self.update(params, grads, alpha)
        max_count = max(1, self.max_count)
        for _ in range(max_count):
            if cost(x) < f0:
                return x
            elif alpha < tol:
                return x
            else:
                alpha = learning_rate * tau
                x = self.update(params, grads, alpha)
                tau *= tau
        return x


class Optimizer:
    r"""Find optimum using gradient-based optimization.

    :param line_search: object that implements algorithm to compute
        search direction :math:`\boldsymbol{s}` given the gradient
        :math:`\nabla_x f` at the current parameter values
        :math:`\boldsymbol{x}` and take multiple steps along it to
        update them according to :math:`\boldsymbol{x} := \boldsymbol{x}
        + \alpha \boldsymbol{s}`
    """

    def __init__(
        self,
        line_search: LineSearch,
    ):  # noqa D107
        self.line_search = line_search
        self.vars_history: Union[list[np.ndarray], None] = None
        self.cost_history: Union[list[np.ndarray], None] = None

    def minimize(
        self,
        x: np.ndarray,
        f: Callable,
        dfdx: Callable,
        alpha: float = 0.01,
        max_iter: int = 100,
        verbose: bool = False,
        epoch: Union[int, None] = None,
        batch: Union[int, None] = None,
        epsilon_absolute: float = 1e-12,
        epsilon_relative: float = 1e-12,
    ) -> np.ndarray:
        r"""Minimize single objective function.

        :param x: parameters to be updated, array of shape (n,)
        :param f: cost function :math:`y = f(\boldsymbol{x})`
        :param alpha: learning rate :math:`\boldsymbol{x} :=
            \boldsymbol{x} + \alpha \boldsymbol{s}`
        :param max_iter: maximum number of optimizer iterations allowed
        :param verbose: whether or not to send progress output to
            standard out
        :param epoch: the epoch in which this optimization is being run
            (for printing)
        :param batch: the batch in which this optimization is being run
            (for printing)
        :param epsilon_absolute: absolute error stopping criterion
        :param epsilon_relative: relative error stopping criterion
        """
        # Stopping criteria (Vanderplaats, "Multidiscipline Design Optimization," ch. 3, p. 121)
        converged = False
        N1 = 0
        N1_max = 100
        N2 = 0
        N2_max = 100

        cost_history: list[np.ndarray] = []
        vars_history: list[np.ndarray] = []

        # Iterative update
        for i in range(0, max_iter):
            y = f(x)

            cost_history.append(y)
            vars_history.append(x)

            x = self.line_search(params=x, cost=f, grads=dfdx(x), learning_rate=alpha)

            if verbose:
                if epoch is not None and batch is not None:
                    e = epoch
                    b = batch
                    print(
                        f"epoch = {e:d}, batch = {b:d}, iter = {i:d}, cost = {y:6.3f}"
                    )
                elif epoch is not None:
                    e = epoch
                    print(f"epoch = {e:d}, iter = {i:d}, cost = {y:6.3f}")
                elif batch is not None:
                    b = batch
                    print(f"batch = {b:d}, iter = {i:d}, cost = {y:6.3f}")
                else:
                    print(f"iter = {i:d}, cost = {y:6.3f}")

            # Absolute convergence criterion
            if i > 1:
                dF1 = abs(cost_history[-1] - cost_history[-2])
                if dF1 < epsilon_absolute * cost_history[0]:
                    N1 += 1
                else:
                    N1 = 0
                if N1 > N1_max:
                    converged = True
                    if verbose:
                        print("Absolute stopping criterion satisfied")

                # Relative convergence criterion
                numerator = abs(cost_history[-1] - cost_history[-2])
                denominator = max(abs(float(cost_history[-1])), 1e-6)
                dF2 = numerator / denominator

                if dF2 < epsilon_relative:
                    N2 += 1
                else:
                    N2 = 0
                if N2 > N2_max:
                    converged = True
                    if verbose:
                        print("Relative stopping criterion satisfied")

                if converged:
                    break

            # Maximum iteration convergence criterion
            if i == max_iter:
                if verbose:
                    print("Maximum optimizer iterations reached")

        self.cost_history = cost_history
        self.vars_history = vars_history

        return x


class GDOptimizer(Optimizer):
    r"""Search for optimum using gradient descent.

    .. warning::
        This optimizer is very inefficient. It was intended
        as a baseline during development. It is not recommended. Use ADAM instead.

    :param tau: amount by which to reduce :math:`\alpha := \tau \times \alpha` on each iteration
    :param tol: stop when cost function doesn't improve more than specified tolerance
    :param max_count: stop when line search iterations exceed maximum count specified
    """

    def __init__(
        self,
        tau: float = 0.5,
        tol: float = 1e-6,
        max_count: int = 1_000,
    ):  # noqa D107
        line_search = Backtracking(
            update=GD(),
            tau=tau,
            tol=tol,
            max_count=max_count,
        )
        super().__init__(line_search)


class ADAMOptimizer(Optimizer):
    r"""Search for optimum using ADAM algorithm.

    :param beta_1: exponential decay rate of 1st moment vector
        :math:`\beta_1\in[0, 1)`
    :param beta_2: exponential decay rate of 2nd moment vector
        :math:`\beta_2\in[0, 1)`
    :param tau: amount by which to reduce :math:`\alpha := \tau \times
        \alpha` on each iteration
    :param tol: stop when cost function doesn't improve more than
        specified tolerance
    :param max_count: stop when line search iterations exceed maximum
        count specified
    """

    def __init__(
        self,
        beta_1: float = 0.9,
        beta_2: float = 0.99,
        tau: float = 0.5,
        tol: float = 1e-12,
        max_count: int = 1_000,
    ):  # noqa D107
        line_search = Backtracking(
            update=ADAM(beta_1, beta_2),
            tau=tau,
            tol=tol,
            max_count=max_count,
        )
        super().__init__(line_search)



"""Test that activation functions are correct."""
import numpy as np
import jenn
from time import time

from importlib.util import find_spec
from ._utils import finite_difference

if find_spec("matplotlib"):
    import matplotlib.pyplot as plt
    MATPLOTLIB_INSTALLED = True
else:
    MATPLOTLIB_INSTALLED = False


ACTIVATIONS = dict(
    relu=jenn.core.activation.Relu,
    tanh=jenn.core.activation.Tanh,
    linear=jenn.core.activation.Linear,
)


def _check_activation_partials(
        x: np.ndarray, activation: jenn.core.activation.Activation):
    """Test 1st and 2nd derivative of activation (default = tanh)"""
    def dy(z):
        """finite difference 1st derivative"""
        return finite_difference(activation.evaluate, z)

    def ddy(z):
        """Finite difference 2nd derivative"""
        return finite_difference(activation.first_derivative, z, dx=1e-6)

    assert np.allclose(activation.first_derivative(x), dy(x), atol=1e-6)
    assert np.allclose(activation.second_derivative(x), ddy(x), atol=1e-6)


def test_inplace():
    """Check that activations update in place."""
    x = np.linspace(-10, 10, 1_000_000)  # some large array to test speed

    y = np.zeros(x.shape)
    dy = np.zeros(x.shape)
    ddy = np.zeros(x.shape)

    id_y = id(y)
    id_dy = id(dy)
    id_ddy = id(ddy)

    for name, activation in ACTIVATIONS.items():

        tic = time()

        y = activation.evaluate(x, y)
        dy = activation.first_derivative(x, y, dy)
        ddy = activation.second_derivative(x, y, dy, ddy)

        assert id(y) == id_y
        assert id(dy) == id_dy
        assert id(ddy) == id_ddy

        toc = time()
        elapsed_time_in_place = toc - tic
        tic = time()

        y_copy = activation.evaluate(x)
        dy_copy = activation.first_derivative(x)
        ddy_copy = activation.second_derivative(x)

        toc = time()
        elapsed_time_copy = toc - tic

        assert id(y_copy) != id_y
        assert id(dy_copy) != id_dy
        assert id(ddy_copy) != id_ddy

        # The other activations are so simple that it's actually faster to
        # not evaluate them in place. Hence, this check only makes sense for
        # TanH, which is the recommended default for JENN applications anyway.
        if name == 'tanh':
            assert elapsed_time_in_place < elapsed_time_copy
        # Speed about "2 x" for TanH: 0.055s < 0.128s for x.shape = (1, 1e6),
        # which translates into significant savings given that TanH gets called
        # for every hidden node in a neural net.


class TestActivation: 
    """Test all activation functions."""

    def test_tanh(self):
        """Test tanh activation"""
        x = np.linspace(-10, 10, 51).reshape((1, -1))
        assert np.allclose(ACTIVATIONS['tanh'].evaluate(x), np.tanh(x), atol=1e-6)
        _check_activation_partials(x, activation=jenn.core.activation.Tanh)

    def test_linear(self):
        """Test linear activation"""
        x = np.linspace(-10, 10, 51).reshape((1, -1))
        assert np.allclose(ACTIVATIONS['linear'].evaluate(x), x, atol=1e-6)
        _check_activation_partials(x, activation=jenn.core.activation.Linear)

    def test_relu(self):
        """Test relu activation"""
        x = np.linspace(-10, 10, 51).reshape((1, -1))
        negative = x <= 0
        positive = x > 0
        array = ACTIVATIONS['relu'].evaluate(x)
        assert np.allclose(array[positive], x[positive], atol=1e-6)
        assert np.allclose(array[negative], 0.0, atol=1e-6)
        _check_activation_partials(
            x[negative].reshape((1, -1)), activation=jenn.core.activation.Relu)
        _check_activation_partials(
            x[positive].reshape((1, -1)), activation=jenn.core.activation.Relu)

    @classmethod
    def plot_activation(cls, name: str):
        """Plot specified activation."""
        if not MATPLOTLIB_INSTALLED:
            raise ValueError(f'Matplotlib is not installed.')

        x = np.linspace(-10, 10, 1_000)
        g = ACTIVATIONS[name]
        y = np.zeros(x.shape)
        dy = np.zeros(x.shape)
        ddy = np.zeros(x.shape)

        y = g.evaluate(x, y)
        dy = g.first_derivative(x, y, dy)
        ddy = g.second_derivative(x, y, dy, ddy)

        plt.plot(x, y)
        plt.title(name + f" (0th derivative)")
        plt.show()

        plt.plot(x, dy)
        plt.title(name + f" (1st derivative)")
        plt.show()

        plt.plot(x, ddy)
        plt.title(name + f" (2nd derivative)")
        plt.show()


if __name__ == "__main__":
    for name in ACTIVATIONS:
        TestActivation.plot_activation(name)



"""Radial Basis Function."""

from typing import Union

import numpy as np


def rbf(
    r: np.ndarray,
    epsilon: float = 0.0,
    out: Union[np.ndarray, None] = None,
) -> np.ndarray:
    """Compute Gaussian Radial Basis Function (RBF).

    :param r: radius from center of RBF
    :param epsilon: hyperparameter
    """
    return np.exp(-((max(0.0, epsilon) * r) ** 2), out=out)
